# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

""""""
import copy
from email.utils import parseaddr

from mailman.email.message import MultipartDigestMessage
from public import public


@public
def copy_headers(from_msg, to_msg, overwrite=False):
    """
    Copy the headers and unixfrom from a message to another one.

    :param from_msg: The source `Message`.
    :type from_msg: email.message.Message
    :param to_msg: The destination `Message`.
    :type to_msg: email.message.Message
    """
    for key, value in from_msg.items():
        if overwrite:
            del to_msg[key]
        if key not in to_msg:
            to_msg[key] = value
    if to_msg.get_unixfrom() is None or overwrite:
        to_msg.set_unixfrom(from_msg.get_unixfrom())
    if (hasattr(from_msg, 'original_size')
        and (getattr(to_msg, 'original_size', None) is None
             or overwrite)):
        to_msg.original_size = from_msg.original_size


@public
def overwrite_message(from_msg, to_msg):
    """
    Overwrite message data of `to_msg` with that of `from_msg`.

    :param from_msg: The source `Message`.
    :type from_msg: email.message.Message
    :param to_msg: The destination `Message`.
    :type to_msg: email.message.Message
    """
    for key in to_msg.keys():
        del to_msg[key]
    for key, value in from_msg.items():
        to_msg[key] = value
    to_msg.set_unixfrom(from_msg.get_unixfrom())
    to_msg.set_payload(from_msg.get_payload(), from_msg.get_charset())
    to_msg.preamble = from_msg.preamble
    to_msg.epilogue = from_msg.epilogue
    to_msg.set_default_type(from_msg.get_default_type())
    if hasattr(from_msg, 'orignal_size'):
        to_msg.original_size = from_msg.original_size


@public
def make_multipart(msg):
    """

    :param msg:
    :type msg: email.message.Message
    :return:
    :rtype: email.message.MIMEMultipart|
            mailman.email.message.MultipartDigestMessage
    """
    if msg.is_multipart():
        out = copy.deepcopy(msg)
    else:
        out = MultipartDigestMessage()
        out.attach(msg)
        copy_headers(msg, out)
    return out


@public
def get_email(msg):
    display_name, email = parseaddr(msg['from'])
    # Address could be None or the empty string.
    if not email:
        email = msg.sender
    return email



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
import unittest
from copy import deepcopy

from mailman.app.lifecycle import create_list
from mailman.interfaces.usermanager import IUserManager
from zope.component import getUtility

from mailman_pgp.config import mm_config
from mailman_pgp.database import mm_transaction, transaction
from mailman_pgp.handlers.signature_strip import SignatureStrip
from mailman_pgp.model.address import PGPAddress
from mailman_pgp.model.list import PGPMailingList
from mailman_pgp.pgp.inline import InlineWrapper
from mailman_pgp.pgp.mime import MIMEWrapper
from mailman_pgp.testing.layers import PGPConfigLayer
from mailman_pgp.testing.pgp import load_key, load_message


class TestSignatureStripHandler(unittest.TestCase):
    layer = PGPConfigLayer

    def setUp(self):
        self.handler = SignatureStrip()

        user_manager = getUtility(IUserManager)
        with mm_transaction():
            self.mlist = create_list('test@example.com',
                                     style_name='pgp-default')
            self.sender = user_manager.create_address('anne@example.com')

        self.pgp_list = PGPMailingList.for_list(self.mlist)
        with transaction():
            self.pgp_list.strip_original_sig = True

        self.sender_key = load_key('rsa_1024.priv.asc')
        with transaction() as t:
            self.pgp_sender = PGPAddress(self.sender)
            self.pgp_sender.key = self.sender_key.pubkey
            self.pgp_sender.key_confirmed = True
            t.add(self.pgp_sender)

        self.msg_clear = load_message('clear.eml')
        self.msg_inline_signed = load_message('inline_signed.eml')
        self.msg_mime_signed = load_message('mime_signed.eml')
        self.msg_inline_signed_invalid = load_message(
                'inline_cleartext_signed_invalid.eml')
        self.msg_mime_signed_invalid = load_message(
                'mime_signed_invalid.eml')

    def test_has_handler(self):
        self.assertIn(SignatureStrip.name, mm_config.handlers.keys())

    def test_no_list(self):
        with mm_transaction():
            ordinary = create_list('ordinary@example.com')

        self.handler.process(ordinary, self.msg_clear, {})

    def test_no_strip(self):
        with transaction():
            self.pgp_list.strip_original_sig = False

        msg = deepcopy(self.msg_mime_signed)
        self.handler.process(self.mlist, msg, {})
        self.assertTrue(MIMEWrapper(msg).is_signed())

        msg = deepcopy(self.msg_inline_signed)
        self.handler.process(self.mlist, msg, {})
        self.assertTrue(InlineWrapper(msg).is_signed())

    def test_strip(self):
        msg = deepcopy(self.msg_mime_signed)
        self.handler.process(self.mlist, msg, {})
        self.assertFalse(MIMEWrapper(msg).is_signed())
        self.assertFalse(MIMEWrapper(msg).has_signature())

        msg = deepcopy(self.msg_inline_signed)
        self.handler.process(self.mlist, msg, {})
        self.assertFalse(InlineWrapper(msg).is_signed())
        self.assertFalse(InlineWrapper(msg).has_signature())



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

"""Test the out-of-process key generator."""
import unittest
from os.path import exists, isfile

from mailman.app.lifecycle import create_list
from parameterized import parameterized
from pgpy.constants import EllipticCurveOID, PubKeyAlgorithm

from mailman_pgp.config import config
from mailman_pgp.database import mm_transaction
from mailman_pgp.model.list import PGPMailingList
from mailman_pgp.pgp.keygen import ListKeyGenerator
from mailman_pgp.testing.layers import PGPConfigLayer
from mailman_pgp.utils.pgp import key_from_file


class TestKeygen(unittest.TestCase):
    layer = PGPConfigLayer

    def setUp(self):
        with mm_transaction():
            self.mlist = create_list('test@example.com',
                                     style_name='pgp-default')
        self.pgp_list = PGPMailingList.for_list(self.mlist)

    @parameterized.expand([
        # RSA + RSA
        (PubKeyAlgorithm.RSAEncryptOrSign, 1024,
         PubKeyAlgorithm.RSAEncryptOrSign, 1024),
        # ECDSA + ECDH
        (PubKeyAlgorithm.ECDSA, EllipticCurveOID.SECP256K1,
         PubKeyAlgorithm.ECDH, EllipticCurveOID.SECP256K1),
        # DSA + ECDH
        (PubKeyAlgorithm.DSA, 1024,
         PubKeyAlgorithm.ECDH, EllipticCurveOID.SECP256K1)
    ])
    def test_generate(self, primary_key_type, primary_key_size, sub_key_type,
                      sub_key_size):
        def reset_primary(primary_key_args):
            config.pgp.primary_key_args = primary_key_args

        self.addCleanup(reset_primary, config.pgp.primary_key_args)

        def reset_sub(sub_key_args):
            config.pgp.sub_key_args = sub_key_args

        self.addCleanup(reset_sub, config.pgp.sub_key_args)

        config.pgp.primary_key_args = (primary_key_type, primary_key_size)
        config.pgp.sub_key_args = (sub_key_type, sub_key_size)

        key_path = self.pgp_list.key_path
        keygen = ListKeyGenerator(self.pgp_list)
        ret_key = keygen.generate(True)
        list_key = self.pgp_list.key
        self.assertTrue(exists(key_path))
        self.assertTrue(isfile(key_path))

        key = key_from_file(key_path)
        self.assertEqual(key.key_algorithm, primary_key_type)
        self.assertEqual(key.key_size, primary_key_size)
        self.assertEqual(ret_key.fingerprint, key.fingerprint)
        self.assertEqual(list_key.fingerprint, key.fingerprint)

        subs = key.subkeys
        self.assertEqual(len(subs), 1)

        keyid, sub = subs.popitem()
        self.assertEqual(sub.key_algorithm, sub_key_type)
        self.assertEqual(sub.key_size, sub_key_size)

        uids = key.userids
        self.assertEqual(len(uids), 2)
        for uid in uids:
            self.assertEqual(uid.name, self.pgp_list.mlist.display_name)
            self.assertIn(uid.email,
                          (self.pgp_list.mlist.posting_address,
                           self.pgp_list.mlist.request_address))



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

"""A PGP plugin for GNU Mailman."""
import warnings

from mailman.interfaces.plugin import IPlugin
from public import public
from zope.interface import implementer

from mailman_pgp.config import config
from mailman_pgp.database import Database
from mailman_pgp.pgp import PGP
from mailman_pgp.rest.root import RESTRoot


@public
@implementer(IPlugin)
class PGPMailman:
    """PGP plugin for Mailman!"""

    def pre_hook(self):
        """See `IPlugin`."""
        # Load the config.
        config.load(self.name)
        config.validate()
        config.convert()
        # Load module global stuff and attach it to the config.
        config.db = Database()
        config.pgp = PGP()
        # Mute PGPy, it gives UserWarnings even during normal use.
        warnings.filterwarnings(action='ignore', category=UserWarning,
                                module='.*pgpy.*')

    def post_hook(self):
        """See `IPlugin`."""
        pass

    def rest_object(self):
        """See `IPlugin`."""
        return RESTRoot()



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

""""""
from functools import wraps

from flufl.lock import Lock


def locked(lockfile, *lock_args, **lock_kwargs):
    """

    :param lockfile:
    :param lock_args:
    :param lock_kwargs:
    :return:
    """

    def locked_decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            with Lock(lockfile, *lock_args, **lock_kwargs):
                return func(*args, **kwargs)

        return wrapper

    return locked_decorator


def locked_obj(lockattr, *lock_args, **lock_kwargs):
    """

    :param lockattr:
    :param lock_args:
    :param lock_kwargs:
    :return:
    """

    def locked_decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            lockfile = getattr(args[0], lockattr)
            locked_func = locked(lockfile, *lock_args, **lock_kwargs)(func)
            return locked_func(*args, **kwargs)

        return wrapper

    return locked_decorator



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

""""""

from mailman.email.message import UserNotification
from mailman.interfaces.pending import IPendings
from mailman.interfaces.subscriptions import TokenOwner
from public import public
from zope.component import getUtility

from mailman_pgp.pgp.mime import MIMEWrapper

SUBSCRIPTION_MOD_REQUEST = """\
----------
TODO: this is a pgp enabled list.
A user with address {address} requested subscription.
The key is attached to this message.

Fingerprint: {fingerprint}
----------
"""

KEY_CHANGE_MOD_REQUEST = """\
----------
TODO: this is a pgp enabled list.
A subscriber with address {address} requested a change of his key.
The new key is attached to this message.

Old key fingerprint: {old_fpr}
New key fingerprint: {new_fpr}
----------
"""

KEY_REVOKE_MOD_REQUEST = """\
----------
TODO: this is a pgp enabled list.
A subscriber with address {address} revoked a part of his key,
which made it unusable and needs to be reset. The subscriber
supplied a new key. The new key is attached to this message.

Old key fingerprint: {old_fpr}
New key fingerprint: {new_fpr}
----------
"""


class ModeratorApprovalMixin:
    def __init__(self, pre_approved=False):
        self.approved = pre_approved

    def _step_mod_approval(self):
        if not self.approved:
            self.push('get_approval')

    def _step_get_approval(self):
        self._pend(TokenOwner.moderator)
        self.push('receive_mod_confirmation')
        self.save()

        name = self._request_name
        body = self._request_body

        if self.mlist.admin_immed_notify:
            subject = 'New {} request from {}'.format(name,
                                                      self.pgp_address.email)
            msg = UserNotification(
                    self.mlist.owner_address, self.mlist.owner_address,
                    subject, body, self.mlist.preferred_language)
            MIMEWrapper(msg).attach_keys(self.pubkey)
            msg.send(self.mlist)
        raise StopIteration

    def _step_receive_mod_confirmation(self):
        pendings = getUtility(IPendings)
        if self.token is not None:
            pendings.confirm(self.token)
            self.token = None
            self.token_owner = TokenOwner.no_one


@public
class ModeratorSubApprovalMixin(ModeratorApprovalMixin):
    def __init__(self, pre_approved=False):
        super().__init__(pre_approved)

    @property
    def _request_name(self):
        return 'subscription'

    @property
    def _request_body(self):
        params = {'mlist': self.mlist.fqdn_listname,
                  'address': self.pgp_address.email,
                  'fingerprint': self.pubkey.fingerprint}
        return SUBSCRIPTION_MOD_REQUEST.format(**params)


@public
class ModeratorKeyChangeApprovalMixin(ModeratorApprovalMixin):
    def __init__(self, pre_approved=False):
        super().__init__(pre_approved)

    @property
    def _request_name(self):
        return 'key change'

    @property
    def _request_body(self):
        params = {'mlist': self.mlist.fqdn_listname,
                  'address': self.pgp_address.email,
                  'fingerprint': self.pubkey.fingerprint,
                  'old_fpr': self.pgp_address.key_fingerprint,
                  'new_fpr': self.pubkey.fingerprint}
        return KEY_CHANGE_MOD_REQUEST.format(**params)


@public
class ModeratorKeyRevokeApprovalMixin(ModeratorApprovalMixin):
    def __init__(self, pre_approved=False):
        super().__init__(pre_approved)

    @property
    def _request_name(self):
        return 'key reset'

    @property
    def _request_body(self):
        params = {'mlist': self.mlist.fqdn_listname,
                  'address': self.pgp_address.email,
                  'fingerprint': self.pubkey.fingerprint,
                  'old_fpr': self.pgp_address.key_fingerprint,
                  'new_fpr': self.pubkey.fingerprint}
        return KEY_REVOKE_MOD_REQUEST.format(**params)



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

""""""
import unittest

from mailman_pgp.config.config import Config
from mailman_pgp.config.validator import ConfigValidator
from mailman_pgp.testing.layers import PGPLayer


class TestValidator(unittest.TestCase):
    layer = PGPLayer

    def test_missing_section(self):
        schema = """\
        [test]
        test_option: something
        """
        validator = ConfigValidator(schema)
        cfg = Config()
        cfg.read_string("""\
        [other_section]
        some_option: something else
        """)
        self.assertRaises(ValueError, validator.validate, cfg)

    def test_additional_section(self):
        schema = """\
        [test]
        test_option: something
        """
        validator = ConfigValidator(schema)
        cfg = Config()
        cfg.read_string("""\
        [test]
        test_option: something
        [other_section]
        some_option: something else
        """)
        self.assertRaises(ValueError, validator.validate, cfg)

    def test_missing_option(self):
        schema = """\
        [test]
        test_option: something
        """
        validator = ConfigValidator(schema)
        cfg = Config()
        cfg.read_string("""\
        [test]
        """)
        self.assertRaises(ValueError, validator.validate, cfg)

    def test_additional_option(self):
        schema = """\
        [test]
        test_option: something
        """
        validator = ConfigValidator(schema)
        cfg = Config()
        cfg.read_string("""\
        [test]
        test_option: something
        other_option: something else
        """)
        self.assertRaises(ValueError, validator.validate, cfg)



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

"""Model for PGP enabled mailing lists."""
from os import system

from mailman.database.types import Enum, SAUnicode
from mailman.interfaces.action import Action
from mailman.interfaces.listmanager import IListManager, ListDeletingEvent
from mailman.interfaces.member import MemberRole
from public import public
from sqlalchemy import Boolean, Column, Integer
from sqlalchemy.orm import reconstructor
from zope.component import getUtility
from zope.event import classhandler

from mailman_pgp.config import config
from mailman_pgp.database import transaction
from mailman_pgp.database.types import EnumFlag
from mailman_pgp.model.base import Base
from mailman_pgp.model.fs_key import FSKey


@public
class PGPMailingList(Base):
    """A PGP enabled mailing list."""

    __tablename__ = 'pgp_lists'

    id = Column(Integer, primary_key=True)
    list_id = Column(SAUnicode, index=True, unique=True)

    # Signature related properties
    unsigned_msg_action = Column(Enum(Action), default=Action.reject)
    inline_pgp_action = Column(Enum(Action), default=Action.defer)
    expired_sig_action = Column(Enum(Action), default=Action.reject)
    revoked_sig_action = Column(Enum(Action), default=Action.reject)
    invalid_sig_action = Column(Enum(Action), default=Action.reject)
    duplicate_sig_action = Column(Enum(Action), default=Action.reject)
    strip_original_sig = Column(Boolean, default=False)
    sign_outgoing = Column(Boolean, default=False)

    # Encryption related properties
    nonencrypted_msg_action = Column(Enum(Action), default=Action.reject)
    encrypt_outgoing = Column(Boolean, default=True)

    # Key related properties
    key_change_workflow = Column(SAUnicode,
                                 default='pgp-key-change-mod-workflow')
    key_signing_allowed = Column(EnumFlag(MemberRole),
                                 default={MemberRole.owner,
                                          MemberRole.moderator})

    def __init__(self, mlist):
        """

        :param mlist:
        :type mlist: mailman.model.mailinglist.MailingList
        """
        super().__init__(list_id=mlist.list_id)
        self._init()
        self._mlist = mlist

    @reconstructor
    def _init(self):
        self._mlist = None
        self._key = FSKey(config.pgp.keydirs['list_keydir'],
                          self.list_id + '.asc', True)

    @property
    def mlist(self):
        """

        :return:
        :rtype: mailman.model.mailinglist.MailingList
        """
        if self._mlist is None:
            self._mlist = getUtility(IListManager).get_by_list_id(self.list_id)
        return self._mlist

    @property
    def fs_key(self):
        return self._key

    @property
    def key(self):
        """
        The private part of the list's keypair.

        :return:
        :rtype: pgpy.PGPKey
        """
        self._key.reload()
        return self._key.key

    @key.setter
    def key(self, value):
        """

        :param value:
        :type value:
        """
        self._key.key = value
        self._key.save()

    @property
    def pubkey(self):
        """
        The public part of the list's keypair.

        :return:
        :rtype: pgpy.PGPKey
        """
        if self.key is None:
            return None
        return self.key.pubkey

    @property
    def key_path(self):
        """
        The path to this list's key in the `list_keydir`.

        :return: List key path.
        :rtype: str
        """
        return self._key.key_path

    @staticmethod
    def for_list(mlist):
        """

        :param mlist:
        :type mlist: mailman.model.mailinglist.MailingList
        :return:
        :rtype: PGPMailingList|None
        """
        if mlist is None:
            return None
        return PGPMailingList.query().filter_by(list_id=mlist.list_id).first()


@classhandler.handler(ListDeletingEvent)
def on_delete(event):
    shred = config.get_value('keypairs', 'shred')
    shred_command = config.get_value('keypairs', 'shred_command')
    delete = config.get_value('keypairs', 'delete')
    pgp_list = PGPMailingList.for_list(event.mailing_list)
    if pgp_list:
        with transaction() as session:
            if delete:
                if shred:
                    if shred_command:
                        system(shred_command + ' ' + pgp_list.key_path)
                    else:
                        pgp_list.fs_key.shred()
                else:
                    pgp_list.fs_key.delete()
            session.delete(pgp_list)



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

""""""

from mailman.core.i18n import _
from mailman.interfaces.workflows import ISubscriptionWorkflow
from mailman.workflows.common import (ConfirmationMixin, VerificationMixin)
from public import public
from zope.interface import implementer

from mailman_pgp.workflows.base import PGPMixin, PGPSubscriptionBase
from mailman_pgp.workflows.key_confirm import ConfirmPubkeyMixin
from mailman_pgp.workflows.key_set import SetPubkeyMixin
from mailman_pgp.workflows.mod_approval import ModeratorSubApprovalMixin


@public
@implementer(ISubscriptionWorkflow)
class OpenSubscriptionPolicy(PGPSubscriptionBase, VerificationMixin,
                             SetPubkeyMixin, ConfirmPubkeyMixin,
                             PGPMixin):
    """"""

    name = 'pgp-policy-open'
    description = _('An open subscription policy, '
                    'for a PGP-enabled mailing list.')
    initial_state = 'prepare'
    save_attributes = (
        'verified',
        'pubkey_key',
        'pubkey_confirmed',
        'address_key',
        'subscriber_key',
        'user_key',
        'token_owner_key',
    )

    def __init__(self, mlist, subscriber=None, *,
                 pre_verified=False, pubkey=None,
                 pubkey_pre_confirmed=False):
        PGPSubscriptionBase.__init__(self, mlist, subscriber)
        VerificationMixin.__init__(self, pre_verified=pre_verified)
        SetPubkeyMixin.__init__(self, pubkey=pubkey)
        ConfirmPubkeyMixin.__init__(self, pre_confirmed=pubkey_pre_confirmed)
        PGPMixin.__init__(self, mlist)

    def _step_prepare(self):
        self.push('do_subscription')
        self.push('restore_subscriber')
        self.push('pubkey_confirmation')
        self.push('restore_address')
        self.push('pubkey_checks')
        self.push('create_address')
        self.push('verification_checks')
        self.push('sanity_checks')


@public
@implementer(ISubscriptionWorkflow)
class ConfirmSubscriptionPolicy(PGPSubscriptionBase, VerificationMixin,
                                ConfirmationMixin, SetPubkeyMixin,
                                ConfirmPubkeyMixin):
    """"""

    name = 'pgp-policy-confirm'
    description = _('A subscription policy, for a PGP-enabled mailing list '
                    'that requires confirmation.')
    initial_state = 'prepare'
    save_attributes = (
        'verified',
        'confirmed',
        'pubkey_key',
        'pubkey_confirmed',
        'address_key',
        'subscriber_key',
        'user_key',
        'token_owner_key',
    )

    def __init__(self, mlist, subscriber=None, *,
                 pre_verified=False, pre_confirmed=False, pubkey=None,
                 pubkey_pre_confirmed=False):
        PGPSubscriptionBase.__init__(self, mlist, subscriber)
        VerificationMixin.__init__(self, pre_verified=pre_verified)
        ConfirmationMixin.__init__(self, pre_confirmed=pre_confirmed)
        SetPubkeyMixin.__init__(self, pubkey=pubkey)
        ConfirmPubkeyMixin.__init__(self, pre_confirmed=pubkey_pre_confirmed)

    def _step_prepare(self):
        self.push('do_subscription')
        self.push('restore_subscriber')
        self.push('pubkey_confirmation')
        self.push('restore_address')
        self.push('pubkey_checks')
        self.push('create_address')
        self.push('confirmation_checks')
        self.push('verification_checks')
        self.push('sanity_checks')


@public
@implementer(ISubscriptionWorkflow)
class ModerationSubscriptionPolicy(PGPSubscriptionBase, VerificationMixin,
                                   ModeratorSubApprovalMixin, SetPubkeyMixin,
                                   ConfirmPubkeyMixin):
    """"""

    name = 'pgp-policy-moderate'
    description = _('A subscription policy, for a PGP-enabled mailing list '
                    'that requires moderation.')
    initial_state = 'prepare'
    save_attributes = (
        'verified',
        'approved',
        'pubkey_key',
        'pubkey_confirmed',
        'address_key',
        'subscriber_key',
        'user_key',
        'token_owner_key',
    )

    def __init__(self, mlist, subscriber=None, *,
                 pre_verified=False, pre_approved=False, pubkey=None,
                 pubkey_pre_confirmed=False):
        PGPSubscriptionBase.__init__(self, mlist, subscriber)
        VerificationMixin.__init__(self, pre_verified=pre_verified)
        ModeratorSubApprovalMixin.__init__(self, pre_approved=pre_approved)
        SetPubkeyMixin.__init__(self, pubkey=pubkey)
        ConfirmPubkeyMixin.__init__(self, pre_confirmed=pubkey_pre_confirmed)

    def _step_prepare(self):
        self.push('do_subscription')
        self.push('restore_subscriber')
        self.push('mod_approval')
        self.push('pubkey_confirmation')
        self.push('restore_address')
        self.push('pubkey_checks')
        self.push('create_address')
        self.push('verification_checks')
        self.push('sanity_checks')


@public
@implementer(ISubscriptionWorkflow)
class ConfirmModerationSubscriptionPolicy(PGPSubscriptionBase,
                                          VerificationMixin,
                                          ConfirmationMixin,
                                          ModeratorSubApprovalMixin,
                                          SetPubkeyMixin, ConfirmPubkeyMixin):
    """"""

    name = 'pgp-policy-confirm-moderate'
    description = _('A subscription policy, for a PGP-enabled mailing list '
                    'that requires moderation after confirmation.')
    initial_state = 'prepare'
    save_attributes = (
        'verified',
        'confirmed',
        'approved',
        'pubkey_key',
        'pubkey_confirmed',
        'address_key',
        'subscriber_key',
        'user_key',
        'token_owner_key',
    )

    def __init__(self, mlist, subscriber=None, *,
                 pre_verified=False, pre_confirmed=False, pre_approved=False,
                 pubkey=None, pubkey_pre_confirmed=False):
        PGPSubscriptionBase.__init__(self, mlist, subscriber)
        VerificationMixin.__init__(self, pre_verified=pre_verified)
        ConfirmationMixin.__init__(self, pre_confirmed=pre_confirmed)
        ModeratorSubApprovalMixin.__init__(self, pre_approved=pre_approved)
        SetPubkeyMixin.__init__(self, pubkey=pubkey)
        ConfirmPubkeyMixin.__init__(self, pre_confirmed=pubkey_pre_confirmed)

    def _step_prepare(self):
        self.push('do_subscription')
        self.push('restore_subscriber')
        self.push('mod_approval')
        self.push('restore_address')
        self.push('pubkey_confirmation')
        self.push('restore_address')
        self.push('pubkey_checks')
        self.push('create_address')
        self.push('confirmation_checks')
        self.push('verification_checks')
        self.push('sanity_checks')



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

"""A PGP enabled posting chain."""

from mailman.chains.base import Link
from mailman.core.i18n import _
from mailman.interfaces.chain import IChain, LinkAction
from public import public
from zope.interface import implementer


@public
@implementer(IChain)
class PGPChain:
    """Default PGP chain."""

    name = 'pgp-posting-chain'
    description = _('The PGP enabled moderation chain.')

    _link_descriptions = (
        ('pgp-encryption', LinkAction.jump, 'pgp-moderation'),
        ('pgp-signature', LinkAction.jump, 'pgp-moderation'),
        ('pgp-mark', LinkAction.defer, None),
        ('truth', LinkAction.jump, 'default-posting-chain')
    )

    def __init__(self):
        self._cached_links = None

    def get_links(self, mlist, msg, msgdata):
        """See `IChain`."""
        if self._cached_links is None:
            self._cached_links = links = []
            for rule, action, chain in self._link_descriptions:
                links.append(Link(rule, action, chain))
        return iter(self._cached_links)



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

""""""
from mailman.email.message import UserNotification
from mailman.interfaces.subscriptions import TokenOwner
from pgpy import PGPKey
from public import public

from mailman_pgp.database import transaction

KEY_REQUEST = """\
----------
TODO: this is a pgp enabled list.
We need your pubkey.
Reply to this message with it as a PGP/MIME(preferred) or inline.
----------"""


@public
class SetPubkeyMixin:
    def __init__(self, pubkey=None):
        self.pubkey = pubkey

    @property
    def pubkey_key(self):
        if self.pubkey is None:
            return None
        return str(self.pubkey)

    @pubkey_key.setter
    def pubkey_key(self, value):
        if value is not None:
            self.pubkey, _ = PGPKey.from_blob(value)
        else:
            self.pubkey = None

    def _step_pubkey_checks(self):
        assert self.pgp_address is not None

        if self.pubkey is None:
            if self.pgp_address.key is None:
                self.push('send_key_request')
        else:
            with transaction():
                self.pgp_address.key = self.pubkey

    def _step_send_key_request(self):
        self._set_token(TokenOwner.subscriber)
        self.push('receive_key')
        self.save()
        request_address = self.mlist.request_address
        email_address = self.address.email
        msg = UserNotification(email_address, request_address,
                               'key set {}'.format(self.token),
                               KEY_REQUEST)
        msg.send(self.mlist, add_precedence=False)
        # Now we wait for the confirmation.
        raise StopIteration

    def _step_receive_key(self):
        self._set_token(TokenOwner.no_one)



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

""""""
import unittest

from mailman_pgp.config.config import Config
from mailman_pgp.config.converter import ConfigConverter
from mailman_pgp.testing.layers import PGPLayer


class TestConverter(unittest.TestCase):
    layer = PGPLayer

    def test_builtins(self):
        schema = """\
        [test]
        test_option: int
        """
        converter = ConfigConverter(schema)
        valid = Config()
        valid.read_string("""\
        [test]
        test_option: 5
        """)
        invalid = Config()
        invalid.read_string("""\
        [test]
        test_option: xyz
        """)
        out = converter.convert(valid)
        self.assertEqual(out['test']['test_option'], 5)
        self.assertRaises(ValueError, converter.convert, invalid)

    def test_callable(self):
        schema = """\
        [test]
        test_option: lazr.config.as_boolean
        """
        converter = ConfigConverter(schema)
        valid = Config()
        valid.read_string("""\
        [test]
        test_option: yes
        """)
        invalid = Config()
        invalid.read_string("""\
        [test]
        test_option: xyz
        """)
        out = converter.convert(valid)
        self.assertEqual(out['test']['test_option'], True)
        self.assertRaises(ValueError, converter.convert, invalid)

    def test_regex(self):
        schema = """\
        [test]
        test_option: 29*4
        """
        converter = ConfigConverter(schema)
        valid = Config()
        valid.read_string("""\
        [test]
        test_option: 2999999994
        """)
        invalid = Config()
        invalid.read_string("""\
        [test]
        test_option: xyz
        """)
        out = converter.convert(valid)
        self.assertEqual(out['test']['test_option'], '2999999994')
        self.assertRaises(ValueError, converter.convert, invalid)

    def test_none(self):
        schema = """\
        [test]
        test_option:
        """
        converter = ConfigConverter(schema)
        cfg = Config()
        cfg.read_string("""\
        [test]
        test_option: something
        """)
        self.assertRaises(ValueError, converter.convert, cfg)



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

"""Miscellaneous PGP utilities."""
from pgpy import PGPKey, PGPSignature
from pgpy.constants import SignatureType
from pgpy.errors import PGPError
from pgpy.packet import Packet, Signature
from pgpy.types import Armorable
from public import public


@public
def expired(verifications):
    """

    :param verifications:
    :return:
    """
    return any(any(sigsubj.signature.is_expired or sigsubj.by.is_expired
                   for sigsubj in verification.good_signatures)
               for verification in verifications)


@public
def revoked(verifications):
    """

    :param verifications:
    :return:
    """
    return any(any(key_revoked(sigsubj.by)
                   for sigsubj in verification.good_signatures)
               for verification in verifications)


@public
def verifies(verifications):
    """

    :param verifications:
    :type verifications: typing.Sequence[pgpy.types.SignatureVerification]
    :return: bool
    """
    return all(bool(verification) and
               all(not sigsubj.signature.is_expired
                   for sigsubj in verification.good_signatures)
               for verification in verifications)


@public
def hashes(verifications):
    """

    :param verifications:
    :return:
    :rtype: typing.Generator[bytes]
    """
    for verification in verifications:
        for sigsubj in verification.good_signatures:
            data = sigsubj.signature.hashdata(sigsubj.subject)
            hasher = sigsubj.signature.hash_algorithm.hasher
            hasher.update(data)
            yield hasher.digest()


@public
def key_from_blob(blob):
    """

    :param blob:
    :return:
    :rtype: pgpy.PGPKey
    """
    key, _ = PGPKey.from_blob(blob)
    return key


@public
def key_from_file(file):
    """

    :param file:
    :return:
    :rtype: pgpy.PGPKey
    """
    key, _ = PGPKey.from_file(file)
    return key


@public
def revoc_from_blob(blob):
    """
    Load a key revocation signature from an ASCII-Armored blob.

    :param blob:
    :return:
    :rtype: pgpy.PGPSignature
    """
    dearm = Armorable.ascii_unarmor(blob)
    p = Packet(dearm['body'])

    if not isinstance(p, Signature):
        raise ValueError('Not a key revocation signature.')
    if p.sigtype not in (SignatureType.KeyRevocation,
                         SignatureType.SubkeyRevocation):
        raise ValueError('Not a key revocation.')

    sig = PGPSignature()
    sig |= p
    return sig


@public
def key_revoked(key):
    """

    :param key:
    :type key: pgpy.PGPKey
    :return:
    :rtype: bool
    """
    if key.is_primary:
        verifier = key
    else:
        verifier = key.parent

    for revoc in key.revocation_signatures:
        try:
            verified = verifier.verify(key, revoc)
        except PGPError:
            continue
        if bool(verified):
            return True

    return False


@public
def key_flags(key):
    """

    :param key:
    :type key: pgpy.PGPKey
    :return:
    :rtype: Set[pgpy.constants.KeyFlags]
    """
    if key.is_expired:
        return set()
    if key_revoked(key):
        return set()

    usage_flags = set()
    uids = (uid for uid in key.userids if uid.is_primary)
    uids = list(uids)
    if len(uids) == 0:
        uids = key.userids

    for uid in uids:
        revoked = False
        for sig in uid.signatures:
            if sig.type is not SignatureType.CertRevocation:
                continue
            if sig.signer == key.fingerprint.keyid:
                try:
                    verified = key.verify(uid, sig)
                except PGPError:
                    continue
                if bool(verified):
                    revoked = True
        if not revoked:
            usage_flags |= uid.selfsig.key_flags
            break

    for subkey in key.subkeys.values():
        if subkey.is_expired:
            continue

        if not key_revoked(subkey):
            usage_flags |= subkey.usage_flags()
    return usage_flags


@public
def key_usable(key, flags_required):
    """
    Check that the `key` has the `flags_required` set of KeyFlags.

    Checks only non-expired, non-revoked key/subkeys. Validates revocations it
    can, so not those made with some other designated revocation key.

    :param key: The key to check.
    :type key: pgpy.PGPKey
    :param flags_required: The set of flags required.
    :type flags_required: set
    :return: Whether the key has the flags_required.
    :rtype: bool
    """
    if key.is_expired:
        return False
    if key_revoked(key):
        return False
    return flags_required.issubset(key_flags(key))


@public
def key_merge(privkey, new_key, signer_key=None):
    """

    :param privkey:
    :type privkey: pgpy.PGPKey
    :param new_key:
    :type new_key: pgpy.PGPKey
    :param signer_key:
    :type signer_key: pgpy.PGPKey
    """
    if privkey.pubkey.key_material != new_key.key_material:
        raise ValueError('You sent a wrong key.')

    uid_map = {}
    for uid in privkey.userids:
        for uid_other in new_key.userids:
            if uid == uid_other:
                uid_map[uid] = uid_other

    if len(uid_map) == 0:
        raise ValueError('No signed UIDs found.')

    uid_sigs = {}
    for uid, uid_other in uid_map.items():
        for sig in uid_other.signatures:
            if sig in uid.signatures:
                continue
            if signer_key is None:
                uid_sigs.setdefault(uid, []).append(sig)
                continue
            if sig.signer != signer_key.fingerprint.keyid:
                continue
            # sig is a new signature, not currently on uid, and seems to
            # be made by the signer_key
            try:
                verification = signer_key.verify(uid, sig)
                if bool(verification):
                    uid_sigs.setdefault(uid, []).append(sig)
            except PGPError:
                pass

    if len(uid_sigs) == 0:
        raise ValueError('No new certifications found.')

    for uid, sigs in uid_sigs.items():
        for sig in sigs:
            uid |= sig



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

""""""
from mailman.rest.helpers import etag, forbidden, not_found, okay

from mailman_pgp.config import config


class AConfig:
    """"""

    def __init__(self, section=None, option=None):
        self._section = section
        self._option = option

    def on_get(self, request, response):
        if not config.get_value('rest', 'allow_read_config'):
            forbidden(response)
            return
        if self._section is None:
            # return whole config
            resource = {}
            for section in config.sections():
                resource[section] = {key: config.get(section, key)
                                     for key in config.options(section)}
        else:
            if not config.has_section(self._section):
                not_found(response)
                return

            if self._option is None:
                # return section
                resource = {key: config.get(self._section, key)
                            for key in config.options(self._section)}
            else:
                if not config.has_option(self._section, self._option):
                    not_found(response)
                    return
                # return value
                resource = {self._option: config.get(self._section,
                                                     self._option)}
        okay(response, etag(resource))



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

"""Mailman PGP configuration module."""
import logging
from configparser import ConfigParser

from mailman.config import config as mailman_config
from mailman.utilities.modules import expand_path
from pkg_resources import resource_string
from public.public import public

from mailman_pgp.config.converter import ConfigConverter
from mailman_pgp.config.validator import ConfigValidator

log = logging.getLogger('mailman.plugin.pgp.config')


@public
class Config(ConfigParser):
    """A ConfigParser with a name."""

    def __init__(self):
        super().__init__()
        self.name = None
        self.dict = None
        schema = resource_string('mailman_pgp.config',
                                 'schema.cfg').decode('utf-8')
        self.validator = ConfigValidator(schema)
        self.converter = ConfigConverter(schema)

    def load(self, name):
        """
        Load the plugin configuration, and set our name.

        :param name: The name to set/load configuration for.
        :type name: str
        """
        self.name = name
        self.read(expand_path(
                dict(mailman_config.plugin_configs)[self.name].configuration))
        log.debug('Config loaded.')

    def validate(self):
        try:
            self.validator.validate(self)
        except ValueError:
            log.exception('Config did not validate.')
            raise
        log.debug('Config validated.')

    def convert(self):
        self.dict = self.converter.convert(self)
        log.debug('Config converted.')

    def get_value(self, section, option):
        return self.dict[section][option]

    def set(self, section, option, value=None):
        self.dict[section][option] = self.converter.converter(section, option)(
                value)
        return super().set(section, option, value)



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

"""The encryption-aware incoming runner."""
import logging

from mailman.config import config as mailman_config
from mailman.core.runner import Runner
from mailman.email.message import Message
from mailman.interfaces.action import Action
from mailman.model.mailinglist import MailingList
from pgpy.errors import PGPError
from public import public

from mailman_pgp.config import config
from mailman_pgp.model.list import PGPMailingList
from mailman_pgp.pgp.wrapper import PGPWrapper
from mailman_pgp.utils.moderation import record_action

log = logging.getLogger('mailman.plugin.pgp')


def _pass_default(msg, msgdata, listid):
    inq = config.get_value('queues', 'in')
    mailman_config.switchboards[inq].enqueue(msg, msgdata,
                                             listid=listid)


@public
class PGPIncomingRunner(Runner):
    def _dispose(self, mlist: MailingList, msg: Message, msgdata: dict):
        """See `IRunner`."""
        # Is the message for an encrypted mailing list? If not, pass to default
        # incoming runner. If yes, go on.
        pgp_list = PGPMailingList.for_list(mlist)
        if not pgp_list:
            _pass_default(msg, msgdata, mlist.list_id)
            return False

        wrapped = PGPWrapper(msg)
        # Is the message encrypted?
        if wrapped.is_encrypted():
            # Decrypt it and pass it on.
            list_key = pgp_list.key
            if list_key is None:
                # keep the message and hope the key becomes available.
                return True

            try:
                msg = wrapped.decrypt(list_key).msg
            except PGPError:
                reason = 'Message could not be decrypted.'
                log.info('[pgp] {}{}: {}'.format(
                        Action.reject.name, msg.get('message-id', 'n/a'),
                        reason))
                record_action(msg, msgdata, Action.reject, msg.sender,
                              reason)
                msgdata['pgp_moderate'] = True
        else:
            # Take the `nonencrypted_msg_action`
            # just set some data for our `encryption` rule which will
            # jump to the moderation chain if `pgp_moderate` is True
            action = pgp_list.nonencrypted_msg_action
            if action != Action.defer:
                reason = 'Message was not encrypted.'
                log.info('[pgp] {}{}: {}'.format(
                        action.name, msg.get('message-id', 'n/a'), reason))
                record_action(msg, msgdata, action, msg.sender, reason)
                msgdata['pgp_moderate'] = True

        _pass_default(msg, msgdata, mlist.list_id)
        return False



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Archives messages to a mbox, locally, encrypted (TBD how),
similar to Mailman's prototype archiver.
"""
import os
from mailbox import mbox

from flufl.lock import Lock
from mailman.interfaces.archiver import IArchiver
from public import public
from zope.interface import implementer

from mailman_pgp.config import config, mm_config
from mailman_pgp.model.list import PGPMailingList
from mailman_pgp.pgp.mime import MIMEWrapper


@public
@implementer(IArchiver)
class LocalMailboxArchiver:
    """Local PGP enabled archiver."""

    name = 'pgp-mbox-local'
    is_enabled = False

    @staticmethod
    def list_url(mlist):
        """See `IArchiver`."""
        return None

    @staticmethod
    def permalink(mlist, msg):
        """See `IArchiver`."""
        return None

    @staticmethod
    def archive_message(mlist, msg):
        """See `IArchiver`."""
        pgp_list = PGPMailingList.for_list(mlist)
        if not pgp_list:
            return None
        mailbox_dir = config.get_value('archiving', 'mailbox_dir')
        mailbox_dir.mkdir(parents=True, exist_ok=True)

        list_dir = mailbox_dir.joinpath(mlist.fqdn_listname)
        mailbox = mbox(str(list_dir))
        lock_file = os.path.join(mm_config.LOCK_DIR,
                                 '{}-{}.lock'.format(mlist.fqdn_listname,
                                                     LocalMailboxArchiver.name)
                                 )
        MIMEWrapper(msg).encrypt(pgp_list.pubkey)
        with Lock(lock_file):
            mailbox.add(msg)
        return None



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

""""""
from datetime import timedelta

from mailman.interfaces.pending import IPendings
from mailman.interfaces.subscriptions import TokenOwner
from mailman.utilities.datetime import now
from mailman.utilities.modules import abstract_component
from mailman.workflows.common import SubscriptionBase
from public import public
from zope.component import getUtility

from mailman_pgp.database import transaction
from mailman_pgp.model.address import PGPAddress
from mailman_pgp.model.list import PGPMailingList


@public
class PGPMixin:
    def __init__(self, mlist, pgp_address=None):
        self.mlist = mlist
        self.pgp_list = PGPMailingList.for_list(mlist)
        self.pgp_address = pgp_address
        if self.pgp_address is not None:
            self.address = self.pgp_address.address

    @property
    def address_key(self):
        return self.pgp_address.email

    @address_key.setter
    def address_key(self, value):
        self.pgp_address = PGPAddress.for_email(value)
        self.member = self.mlist.regular_members.get_member(value)

    def _step_create_address(self):
        self.pgp_address = PGPAddress.for_address(self.address)
        if self.pgp_address is None:
            with transaction() as t:
                self.pgp_address = PGPAddress(self.address)
                t.add(self.pgp_address)

    def _step_restore_address(self):
        self.pgp_address = PGPAddress.for_address(self.address)

    def _set_token(self, token_owner):
        assert isinstance(token_owner, TokenOwner)
        pendings = getUtility(IPendings)
        if self.token is not None:
            pendings.confirm(self.token)
        self.token_owner = token_owner
        if token_owner is TokenOwner.no_one:
            self.token = None
            return

        pendable = self.pendable_class()(
                list_id=self.mlist.list_id,
                email=self.address.email,
                display_name=self.address.display_name,
                when=now().replace(microsecond=0).isoformat(),
                token_owner=token_owner.name,
        )
        self.token = pendings.add(pendable, timedelta(days=3650))


@public
@abstract_component
class PGPSubscriptionBase(SubscriptionBase, PGPMixin):
    def __init__(self, mlist, subscriber=None, *, pgp_address=None):
        SubscriptionBase.__init__(self, mlist, subscriber)
        PGPMixin.__init__(self, mlist, pgp_address=pgp_address)

    def _step_restore_subscriber(self):
        self._restore_subscriber()



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

""""""

from pgpy import PGPKeyring
from public import public

from mailman_pgp.config import config
from mailman_pgp.utils.config import key_spec


@public
class PGP:
    def __init__(self):
        self.primary_key_args = key_spec(
            config.get_value('keypairs', 'primary_key'))
        self.sub_key_args = key_spec(config.get_value('keypairs', 'sub_key'))
        # Make sure the keydir paths are directories and exist.
        self.keydirs = {keydir_name: config.get_value('keydirs', keydir_name)
                        for keydir_name in config.options('keydirs')}

        for keydir_path in self.keydirs.values():
            # TODO set a strict mode here
            keydir_path.mkdir(parents=True, exist_ok=True)

    def _keyring(self, keydir):
        directory = self.keydirs[keydir]
        return PGPKeyring(*map(str, directory.glob('*.asc')))

    @property
    def list_keyring(self):
        return self._keyring('list_keydir')

    @property
    def user_keyring(self):
        return self._keyring('user_keydir')

    @property
    def archive_keyring(self):
        return self._keyring('archive_keydir')



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
import unittest

from mailman.app.lifecycle import create_list
from mailman.interfaces.action import Action
from mailman.interfaces.member import MemberRole
from mailman.interfaces.usermanager import IUserManager
from mailman.testing.helpers import (get_queue_messages, make_testable_runner,
                                     set_preferred,
                                     specialized_message_from_string as mfs)
from zope.component import getUtility

from mailman_pgp.config import mm_config
from mailman_pgp.database import mm_transaction, transaction
from mailman_pgp.model.address import PGPAddress
from mailman_pgp.model.list import PGPMailingList
from mailman_pgp.pgp.wrapper import PGPWrapper
from mailman_pgp.runners.incoming import PGPIncomingRunner
from mailman_pgp.testing.layers import PGPConfigLayer
from mailman_pgp.testing.pgp import load_key, load_message


class TestPGPIncomingRunner(unittest.TestCase):
    layer = PGPConfigLayer

    def setUp(self):
        user_manager = getUtility(IUserManager)
        with mm_transaction():
            self.mlist = create_list('test@example.com',
                                     style_name='pgp-default')
            self.sender = user_manager.create_user('RSA-1024b@example.org')
            set_preferred(self.sender)
            self.mlist.subscribe(self.sender, MemberRole.member)

        self.list_key = load_key('ecc_p256.priv.asc')
        self.pgp_list = PGPMailingList.for_list(self.mlist)
        self.pgp_list.key = self.list_key

        self.sender_key = load_key('rsa_1024.priv.asc')
        with transaction() as t:
            self.pgp_sender = PGPAddress(self.sender.preferred_address)
            self.pgp_sender.key = self.sender_key.pubkey
            t.add(self.pgp_sender)

        self.msg_clear = load_message('clear.eml')

        self.runner = make_testable_runner(PGPIncomingRunner, 'in')

    def test_pass_default(self):
        with mm_transaction():
            create_list('ordinary@example.com')

        msg = mfs("""\
From: anne@example.com
To: ordinary@example.com

""")

        msgdata = dict(listid='ordinary.example.com')
        mm_config.switchboards['in'].enqueue(msg, msgdata)
        self.runner.run()
        items = get_queue_messages('in_default', expected_count=1)
        self.assertEqual(items[0].msg.sender, 'anne@example.com')

    def test_no_key(self):
        with mm_transaction():
            create_list('no-key@example.com',
                        style_name='pgp-default')
        msg = mfs("""\
From: RSA-1024b@example.org
To: no-key@example.com

Some text.
""")
        PGPWrapper(msg).encrypt(self.pgp_list.pubkey)

        msgdata = dict(listid='no-key.example.com')
        mm_config.switchboards['in'].enqueue(msg, msgdata)
        runner = make_testable_runner(PGPIncomingRunner, 'in',
                                      lambda runner: True)
        runner.run()
        # Expect the message still there. Waiting for list key.
        get_queue_messages('in', expected_count=1)

    def test_nonencrypted_action(self):
        with transaction():
            self.pgp_list.nonencrypted_msg_action = Action.hold

        msgdata = dict(listid='test.example.com')
        mm_config.switchboards['in'].enqueue(self.msg_clear, msgdata)
        self.runner.run()
        items = get_queue_messages('in_default', expected_count=1)
        self.assertEqual(items[0].msgdata['pgp_action'],
                         Action.hold.name)
        self.assertEqual(items[0].msgdata['moderation_sender'],
                         self.msg_clear.sender)
        self.assertEqual(items[0].msgdata['moderation_reasons'],
                         ['Message was not encrypted.'])
        self.assertTrue(items[0].msgdata['pgp_moderate'])

        with transaction():
            self.pgp_list.nonencrypted_msg_action = Action.defer

        msgdata = dict(listid='test.example.com')
        mm_config.switchboards['in'].enqueue(self.msg_clear, msgdata)
        self.runner.run()
        get_queue_messages('in_default', expected_count=1)

    def test_decrypt(self):
        payload = 'Some encrypted text.'
        msg = mfs("""\
From: RSA-1024b@example.org
To: test@example.com

{}
""".format(str(payload)))
        encrypted = PGPWrapper(msg).copy().encrypt(self.pgp_list.pubkey).msg

        msgdata = dict(listid='test.example.com')
        mm_config.switchboards['in'].enqueue(encrypted, msgdata)
        self.runner.run()
        items = get_queue_messages('in_default', expected_count=1)
        out_msg = items[0].msg
        self.assertEqual(out_msg.get_payload(), msg.get_payload())

    def test_decrypt_combined(self):
        payload = 'Some signed and encrypted text.'
        msg = mfs("""\
From: RSA-1024b@example.org
To: test@example.com

{}
""".format(str(payload)))
        PGPWrapper(msg).sign_encrypt(self.sender_key,
                                     self.pgp_list.pubkey,
                                     self.pgp_sender.key)

        msgdata = dict(listid='test.example.com')
        mm_config.switchboards['in'].enqueue(msg, msgdata)

        self.runner.run()
        items = get_queue_messages('in_default', expected_count=1)
        out_msg = items[0].msg
        out_wrapped = PGPWrapper(out_msg)
        self.assertTrue(out_wrapped.is_signed())
        self.assertTrue(out_wrapped.verifies(self.pgp_sender.key))

    def test_decrypt_fail(self):
        payload = 'Some signed and encrypted text.'
        msg = mfs("""\
From: RSA-1024b@example.org
To: test@example.com

{}
        """.format(str(payload)))

        PGPWrapper(msg).encrypt(self.sender_key.pubkey)

        msgdata = dict(listid='test.example.com')
        mm_config.switchboards['in'].enqueue(msg, msgdata)
        self.runner.run()
        items = get_queue_messages('in_default', expected_count=1)
        self.assertEqual(items[0].msgdata['pgp_action'],
                         Action.reject.name)
        self.assertEqual(items[0].msgdata['moderation_sender'],
                         msg.sender)
        self.assertEqual(items[0].msgdata['moderation_reasons'],
                         ['Message could not be decrypted.'])
        self.assertTrue(items[0].msgdata['pgp_moderate'])



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

""""""
import unittest
from urllib.error import HTTPError

from mailman.interfaces.usermanager import IUserManager
from mailman.testing.helpers import call_api
from zope.component import getUtility

from mailman_pgp.database import mm_transaction, transaction
from mailman_pgp.model.address import PGPAddress
from mailman_pgp.testing.layers import PGPRESTLayer
from mailman_pgp.testing.pgp import load_key
from mailman_pgp.utils.pgp import key_from_blob


class TestAddresses(unittest.TestCase):
    layer = PGPRESTLayer

    def setUp(self):
        with mm_transaction():
            self.mm_address = getUtility(IUserManager).create_address(
                    'anne@example.com')
        self.address_key = load_key('ecc_p256.pub.asc')
        with transaction() as t:
            self.address = PGPAddress(self.mm_address)
            self.address.key = self.address_key
            self.address.key_confirmed = True
            t.add(self.address)

    def test_missing_address(self):
        with self.assertRaises(HTTPError) as cm:
            call_api('http://localhost:9001/3.1/plugins/pgp/addresses/'
                     'bart@example.com')
        self.assertEqual(cm.exception.code, 404)

    def test_all_addresses(self):
        json, response = call_api(
                'http://localhost:9001/3.1/plugins/pgp/addresses/')
        self.assertEqual(json['total_size'], 1)
        self.assertEqual(len(json['entries']), 1)
        addresses = json['entries']
        address = addresses[0]
        self.assertEqual(address['email'], self.address.email)

    def test_get_address(self):
        json, response = call_api(
                'http://localhost:9001/3.1/plugins/pgp/addresses/'
                'anne@example.com')
        self.assertEqual(json['email'], self.address.email)

    def test_address_key(self):
        json, response = call_api(
                'http://localhost:9001/3.1/plugins/pgp/addresses/'
                'anne@example.com/key')
        key = key_from_blob(json['key'])
        self.assertEqual(key.fingerprint, self.address_key.fingerprint)
        self.assertEqual(json['key_fingerprint'], self.address_key.fingerprint)
        self.assertEqual(json['key_confirmed'], True)

    def test_address_no_key(self):
        with mm_transaction():
            mm_address = getUtility(IUserManager).create_address(
                    'bart@example.com')
        with transaction() as t:
            address = PGPAddress(mm_address)
            t.add(address)
        with self.assertRaises(HTTPError) as cm:
            call_api('http://localhost:9001/3.1/plugins/pgp/addresses/'
                     'bart@example.com/key')
        self.assertEqual(cm.exception.code, 404)



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

""""""
from mailman.pipelines.base import BasePipeline

from public import public


@public
class PGPPostingPipeline(BasePipeline):
    """The built-in owner pipeline."""

    name = 'pgp-posting-pipeline'
    description = 'PGP posting pipeline'

    _default_handlers = (
        'pgp-signature-strip',
        'mime-delete',
        'tagger',
        'member-recipients',
        'avoid-duplicates',
        'cleanse',
        'cleanse-dkim',
        'cook-headers',
        'subject-prefix',
        'rfc-2369',
        'to-archive',
        'to-digest',
        'to-usenet',
        'after-delivery',
        'acknowledge',
        'decorate',
        'dmarc',
        'to-outgoing',
    )



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

"""Encryption moderation rule for the pgp-posting-chain."""

from mailman.core.i18n import _
from mailman.interfaces.rules import IRule
from public.public import public
from zope.interface import implementer


@public
@implementer(IRule)
class Encryption:
    """The encryption moderation rule."""

    name = 'pgp-encryption'
    description = _(
            "A rule which jumps to the moderation chain, "
            "when the incoming runner instructs it to.")
    record = False

    def check(self, mlist, msg, msgdata):
        """See `IRule`."""
        return msgdata.get('pgp_moderate', False)



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

"""Database stored PGP key."""
from public import public
from sqlalchemy import Column, Integer, LargeBinary
from sqlalchemy.orm import reconstructor

from mailman_pgp.model.base import Base
from mailman_pgp.utils.pgp import key_from_blob


@public
class DBKey(Base):
    """Database stored PGP key."""
    __tablename__ = 'keys'

    id = Column(Integer, primary_key=True)
    _key_material = Column('key_material', LargeBinary)

    def __init__(self, key=None):
        super().__init__()
        self._init()
        self.key = key

    @reconstructor
    def _init(self):
        self._key = None

    @property
    def key(self):
        if self._key is None:
            self._key = key_from_blob(self._key_material)
        return self._key

    @key.setter
    def key(self, value):
        if value is None:
            self._key = None
            self._key_material = bytes()
        else:
            self._key = value
            self._key_material = bytes(value)



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

""""""
from mailman.email.message import UserNotification
from mailman.interfaces.subscriptions import TokenOwner
from public import public

from mailman_pgp.database import transaction
from mailman_pgp.model.list import PGPMailingList
from mailman_pgp.pgp.wrapper import PGPWrapper

CONFIRM_REQUEST = """\
----------
TODO: this is a pgp enabled list.
Reply to this message with this whole text
signed with your supplied key, either inline or PGP/MIME.

Fingerprint: {}
Token: {}
----------
"""


@public
class ConfirmPubkeyMixin:
    def __init__(self, pre_confirmed=False):
        self.pubkey_confirmed = pre_confirmed

    def _step_pubkey_confirmation(self):
        assert self.pgp_address is not None

        if self.pubkey_confirmed:
            with transaction():
                self.pgp_address.key_confirmed = True
        else:
            if not self.pgp_address.key_confirmed:
                self.push('send_key_confirm_request')

    def _step_send_key_confirm_request(self):
        self._set_token(TokenOwner.subscriber)
        self.push('receive_key_confirmation')
        self.save()

        request_address = self.mlist.request_address
        email_address = self.address.email
        msg = UserNotification(email_address, request_address,
                               'key confirm {}'.format(self.token),
                               CONFIRM_REQUEST.format(
                                       self.pgp_address.key_fingerprint,
                                       self.token))
        pgp_list = PGPMailingList.for_list(self.mlist)
        PGPWrapper(msg).sign_encrypt(pgp_list.key, self.pgp_address.key)

        msg.send(self.mlist)
        raise StopIteration

    def _step_receive_key_confirmation(self):
        self._set_token(TokenOwner.no_one)



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

""""""
import unittest

from mailman_pgp.config import config
from mailman_pgp.database import transaction
from mailman_pgp.model.db_key import DBKey
from mailman_pgp.testing.layers import PGPConfigLayer
from mailman_pgp.testing.pgp import load_key


class TestDBKey(unittest.TestCase):
    layer = PGPConfigLayer

    def test_key(self):
        with transaction() as t:
            db_key = DBKey()
            t.add(db_key)

        key_data = load_key('rsa_1024.priv.asc')
        with transaction():
            db_key.key = key_data
        self.assertEqual(db_key.key.fingerprint, key_data.fingerprint)

    def test_loaded(self):
        key_data = load_key('rsa_1024.priv.asc')
        with transaction() as t:
            db_key = DBKey(key_data)
            t.add(db_key)

        config.db.scoped_session.remove()

        loaded_key = DBKey.query().one()
        self.assertEqual(loaded_key.key.fingerprint, key_data.fingerprint)



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

""""""
import pathlib

from mailman.utilities.string import expand
from pgpy.constants import EllipticCurveOID, PubKeyAlgorithm

from mailman_pgp.config import mm_config


def expandable_str(value):
    return expand(value, None, mm_config.paths)


def expandable_path(value):
    return pathlib.Path(expandable_str(value))


def key_spec(value):
    KEYPAIR_TYPE_MAP = {
        'RSA': PubKeyAlgorithm.RSAEncryptOrSign,
        'DSA': PubKeyAlgorithm.DSA,
        'ECDSA': PubKeyAlgorithm.ECDSA,
        'ECDH': PubKeyAlgorithm.ECDH
    }
    ECC_OID_MAP = {
        'nistp256': EllipticCurveOID.NIST_P256,
        'nistp384': EllipticCurveOID.NIST_P384,
        'nistp521': EllipticCurveOID.NIST_P521,
        'brainpoolP256r1': EllipticCurveOID.Brainpool_P256,
        'brainpoolP384r1': EllipticCurveOID.Brainpool_P384,
        'brainpoolP512r1': EllipticCurveOID.Brainpool_P512,
        'secp256k1': EllipticCurveOID.SECP256K1
    }
    key_type, key_length = value.split(':')
    key_type = key_type.upper()
    key_length = key_length.lower()

    if key_type not in KEYPAIR_TYPE_MAP:
        raise ValueError('Invalid key type: {}.'.format(key_type))

    out_type = KEYPAIR_TYPE_MAP[key_type]
    if key_type in ('ECDSA', 'ECDH'):
        if key_length not in ECC_OID_MAP:
            raise ValueError('Invalid key length: {}.'.format(key_length))
        out_length = ECC_OID_MAP[key_length]
    else:
        out_length = int(key_length)
    return (out_type, out_length)



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

""""""
import unittest

from mailman.app.lifecycle import create_list
from mailman.interfaces.usermanager import IUserManager
from zope.component import getUtility

from mailman_pgp.database import mm_transaction
from mailman_pgp.testing.layers import PGPConfigLayer
from mailman_pgp.workflows.subscription import (
    ConfirmModerationSubscriptionPolicy, ConfirmSubscriptionPolicy,
    ModerationSubscriptionPolicy, OpenSubscriptionPolicy)


class TestSubscriptionWorkflows(unittest.TestCase):
    layer = PGPConfigLayer

    def setUp(self):
        with mm_transaction():
            self.mlist = create_list('test@example.com',
                                     style_name='pgp-default')
        self.sender = getUtility(IUserManager).create_address(
                'rsa-1024b@example.org')

    def test_open_policy(self):
        workflow = OpenSubscriptionPolicy(self.mlist, self.sender)
        next(workflow)

    def test_confirm_policy(self):
        workflow = ConfirmSubscriptionPolicy(self.mlist, self.sender)
        next(workflow)

    def test_moderation_policy(self):
        workflow = ModerationSubscriptionPolicy(self.mlist, self.sender)
        next(workflow)

    def test_confirm_moderation_policy(self):
        workflow = ConfirmModerationSubscriptionPolicy(self.mlist, self.sender)
        next(workflow)



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

"""REST interface to a PGP enabled mailing list."""
from lazr.config import as_boolean
from mailman.interfaces.action import Action
from mailman.interfaces.listmanager import IListManager
from mailman.interfaces.member import MemberRole
from mailman.rest.helpers import (accepted, bad_request,
                                  child, CollectionMixin, etag,
                                  forbidden, GetterSetter, no_content,
                                  not_found, NotFound, okay)
from mailman.rest.validator import (enum_validator, PatchValidator,
                                    UnknownPATCHRequestError, Validator)
from pgpy.errors import PGPError
from public import public
from zope.component import getUtility

from mailman_pgp.config import config
from mailman_pgp.database import transaction
from mailman_pgp.model.list import PGPMailingList
from mailman_pgp.utils.pgp import key_from_blob, key_merge
from mailman_pgp.utils.rest import enumflag_validator, workflow_validator
from mailman_pgp.workflows.key_change import (KeyChangeModWorkflow,
                                              KeyChangeWorkflow)

ATTRIBUTES = dict(
        list_id=GetterSetter(None),
        unsigned_msg_action=GetterSetter(enum_validator(Action)),
        inline_pgp_action=GetterSetter(enum_validator(Action)),
        expired_sig_action=GetterSetter(enum_validator(Action)),
        revoked_sig_action=GetterSetter(enum_validator(Action)),
        invalid_sig_action=GetterSetter(enum_validator(Action)),
        duplicate_sig_action=GetterSetter(enum_validator(Action)),
        strip_original_sig=GetterSetter(as_boolean),
        sign_outgoing=GetterSetter(as_boolean),
        nonencrypted_msg_action=GetterSetter(enum_validator(Action)),
        encrypt_outgoing=GetterSetter(as_boolean),
        key_change_workflow=GetterSetter(
                workflow_validator(KeyChangeWorkflow, KeyChangeModWorkflow)),
        key_signing_allowed=GetterSetter(enumflag_validator(MemberRole))
)

VALIDATORS = ATTRIBUTES.copy()
for attribute, gettersetter in list(VALIDATORS.items()):
    if gettersetter.decoder is None:
        del VALIDATORS[attribute]


class _PGPListBase(CollectionMixin):
    def _resource_as_dict(self, emlist):
        """See `CollectionMixin`."""
        resource = {}
        for attribute, getter in ATTRIBUTES.items():
            value = getter.get(emlist, attribute)
            if isinstance(value, (set, frozenset)):
                value = list(iter(value))
            resource[attribute] = value
        resource['self_link'] = self.api.path_to(
                '/plugins/{}/lists/{}'.format(config.name,
                                              emlist.list_id))
        return resource

    def _get_collection(self, request):
        """See `CollectionMixin`."""
        return PGPMailingList.query().all()


@public
class AllPGPLists(_PGPListBase):
    """The PGP enabled mailing lists."""

    def on_get(self, request, response):
        """/lists"""
        resource = self._make_collection(request)
        return okay(response, etag(resource))


@public
class APGPList(_PGPListBase):
    """One PGP enabled mailing list."""

    def __init__(self, list_identifier):
        manager = getUtility(IListManager)
        if '@' in list_identifier:
            mlist = manager.get(list_identifier)
        else:
            mlist = manager.get_by_list_id(list_identifier)
        self._mlist = PGPMailingList.for_list(mlist)

    def on_get(self, request, response):
        """/lists/<list_id>"""
        if self._mlist is None:
            not_found(response)
        else:
            okay(response, self._resource_as_json(self._mlist))

    def on_put(self, request, response):
        """/lists/<list_id>"""
        if self._mlist is None:
            not_found(response)
        else:
            validator = Validator(**VALIDATORS)
            try:
                with transaction():
                    validator.update(self._mlist, request)
            except ValueError as error:
                bad_request(response, str(error))
            else:
                no_content(response)

    def on_patch(self, request, response):
        """/lists/<list_id>"""
        if self._mlist is None:
            not_found(response)
        else:
            try:
                validator = PatchValidator(request, VALIDATORS)
            except UnknownPATCHRequestError as error:
                bad_request(response,
                            'Unknown attribute: {}'.format(error.attribute))
                return
            try:
                with transaction():
                    validator.update(self._mlist, request)
            except ValueError as error:
                bad_request(response, str(error))
            else:
                no_content(response)

    @child()
    def key(self, context, segments):
        if self._mlist is None:
            return NotFound(), []
        return AListKey(self._mlist), []

    @child()
    def pubkey(self, context, segments):
        if self._mlist is None:
            return NotFound(), []
        return AListPubkey(self._mlist), []


@public
class AListKey:
    """A PGP private key."""

    def __init__(self, mlist):
        self._mlist = mlist

    def on_get(self, request, response):
        """/lists/<list_id>/key"""
        if not config.get_value('rest', 'allow_read_private_key'):
            forbidden(response)
            return
        key = self._mlist.key
        if key is None:
            not_found(response)
        else:
            resource = dict(key=str(key),
                            key_fingerprint=str(key.fingerprint))
            okay(response, etag(resource))

    def on_put(self, request, response):
        """/lists/<list_id>/key"""
        if not config.get_value('rest', 'allow_write_private_key'):
            forbidden(response)
            return
        try:
            validator = Validator(key=GetterSetter(key_from_blob))
            values = validator(request)
        except (ValueError, PGPError) as error:
            bad_request(response, str(error))
            return

        key = values.pop('key')
        if key.is_public:
            bad_request(response, 'Cannot set a public key.')
            return

        with transaction():
            self._mlist.key = key
        accepted(response)


@public
class AListPubkey:
    """A PGP list public key."""

    def __init__(self, mlist):
        self._mlist = mlist

    def on_get(self, request, response):
        """/lists/<list_id>/pubkey"""
        pubkey = self._mlist.pubkey
        if pubkey is None:
            not_found(response)
        else:
            resource = dict(public_key=str(pubkey),
                            key_fingerprint=str(pubkey.fingerprint))
            okay(response, etag(resource))

    def on_put(self, request, response):
        """/lists/<list_id>/pubkey"""
        # merge pubkey with the current one. Assume all new sigs are valid.
        try:
            validator = Validator(public_key=GetterSetter(key_from_blob))
            values = validator(request)
        except (ValueError, PGPError) as error:
            bad_request(response, str(error))
            return

        key = values.pop('public_key')
        try:
            key_merge(self._mlist.key, key)
            self._mlist.fs_key.save()
        except ValueError as e:
            bad_request(response, str(e))
            return

        accepted(response)



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
""""""

from mailman.interfaces.handler import IHandler
from public import public
from zope.interface import implementer

from mailman_pgp.model.list import PGPMailingList
from mailman_pgp.pgp.wrapper import PGPWrapper


@public
@implementer(IHandler)
class SignatureStrip:
    name = 'pgp-signature-strip'
    description = 'Strip the signature of the message.'

    def process(self, mlist, msg, msgdata):
        """See `IHandler`."""
        pgp_list = PGPMailingList.for_list(mlist)
        if not pgp_list or not pgp_list.strip_original_sig:
            return

        PGPWrapper(msg).strip_signature()



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

""""""
from public import public


@public
def record_action(msg, msgdata, action, sender, reason):
    """

    :param msg:
    :param msgdata:
    :param action:
    :param sender:
    :param reason:
    :return:
    """
    msgdata['pgp_action'] = action.name
    msgdata['moderation_sender'] = sender
    msgdata.setdefault('moderation_reasons', []).append(reason)



#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

"""A combined PGP/MIME + inline PGP wrapper."""

from pgpy.errors import PGPError
from public import public

from mailman_pgp.pgp.base import BaseWrapper
from mailman_pgp.pgp.inline import InlineWrapper
from mailman_pgp.pgp.mime import MIMEWrapper
from mailman_pgp.pgp.mime_multisig import MIMEMultiSigWrapper
from mailman_pgp.utils.pgp import verifies


@public
class PGPWrapper(BaseWrapper):
    """A combined PGP/MIME + inline PGP wrapper."""

    def __init__(self, msg, copy=False, default=MIMEWrapper):
        """
        Wrap the given message.

        :param msg: The message to wrap.
        :type msg: mailman.email.message.Message
        :param copy: Whether to copy the message when wrapping.
        :type copy: bool
        :param default: The wrapper class used for active operations (sign,
                        encrypt, attach_keys, attach_revocs)
        :type default: Type[MIMEWrapper|MIMEMultiSigWrapper|InlineWrapper]
        """
        super().__init__(msg, copy)
        self.mime = MIMEWrapper(self.msg)
        self.inline = InlineWrapper(self.msg)
        self.multisig = MIMEMultiSigWrapper(self.msg)
        self.wrappers = (self.mime, self.inline, self.multisig)
        if default is MIMEWrapper:
            self.default = self.mime
        elif default is MIMEMultiSigWrapper:
            self.default = self.multisig
        elif default is InlineWrapper:
            self.default = self.inline
        else:
            raise ValueError('Default wrapper must be one of ' +
                             MIMEWrapper.__name__ + ' ' +
                             MIMEMultiSigWrapper.__name__ + ' ' +
                             InlineWrapper.__name__ + '.')

    def _rewrap(self, wrapper):
        if wrapper is not None:
            return PGPWrapper(wrapper.msg, default=self.default.__class__)

    def get_payload(self):
        return self.default.get_payload()

    def is_signed(self):
        """
        Whether this message is signed.

        :return: If the message is signed.
        :rtype: bool
        """
        return any(wrapper.is_signed() for wrapper in self.wrappers)

    def has_signature(self):
        """
        Whether some parts of the message are signed.

        :return: If some parts of the message are signed.
        :rtype: bool
        """
        return any(wrapper.has_signature() for wrapper in self.wrappers)

    def get_signed(self):
        """
        Get the signed content of the message.

        :return: The signed contents of the message.
        :rtype: typing.Generator[str]
        """
        if self.mime.is_signed():
            yield from self.mime.get_signed()
        elif self.multisig.is_signed():
            yield from self.multisig.get_signed()
        elif self.inline.is_signed():
            yield from self.inline.get_signed()

    def get_signature(self):
        """

        :return:
        :rtype: typing.Generator[pgpy.PGPMessage|pgpy.PGPSignature|
                                 pgpy.PGPDetachedSignature]
        """
        if self.mime.is_signed():
            yield from self.mime.get_signature()
        elif self.multisig.is_signed():
            yield from self.multisig.get_signature()
        elif self.inline.is_signed():
            yield from self.inline.get_signature()

    def strip_signature(self):
        """

        :return:
        :rtype: PGPWrapper
        """
        result = None
        if self.mime.is_signed():
            result = self.mime.strip_signature()
        elif self.multisig.is_signed():
            result = self.multisig.strip_signature()
        elif self.inline.is_signed():
            result = self.inline.strip_signature()
        return self._rewrap(result)

    def sign(self, key, **kwargs):
        """
        Sign a message with key.

        :param key: The key to sign with.
        :type key: pgpy.PGPKey
        :return:
        :rtype: PGPWrapper
        """
        return self._rewrap(self.default.sign(key, **kwargs))

    def verify(self, key):
        """
        Verify the signatures of this message with key.

        :param key: The key to verify with.
        :type key: pgpy.PGPKey
        :return: The verified signatures.
        :rtype: typing.Generator[pgpy.types.SignatureVerification]
        """
        if self.mime.is_signed():
            yield from self.mime.verify(key)
        elif self.multisig.is_signed():
            yield from self.multisig.verify(key)
        elif self.inline.is_signed():
            yield from self.inline.verify(key)

    def verifies(self, key):
        return verifies(self.verify(key))

    def is_encrypted(self):
        """
        Whether the message is encrypted.

        :return: If the message is encrypted.
        :rtype: bool
        """
        return any(wrapper.is_encrypted() for wrapper in self.wrappers)

    def has_encryption(self):
        """
        Whether some parts of the message are encrypted.

        :return: If some parts of the message are encrypted.
        :rtype: bool
        """
        return any(wrapper.has_encryption() for wrapper in self.wrappers)

    def get_encrypted(self):
        """

        :return:
        :rtype: typing.Generator[pgpy.PGPMessage]
        """
        if self.mime.is_encrypted():
            yield from self.mime.get_encrypted()
        elif self.multisig.is_encrypted():
            yield from self.mime.get_encrypted()
        elif self.inline.is_encrypted():
            yield from self.inline.get_encrypted()

    def encrypt(self, *keys, **kwargs):
        """
        Encrypt the message with key/s, using cipher.

        :param keys: The key/s to encrypt with.
        :type keys: pgpy.PGPKey
        :return:
        :rtype: PGPWrapper
        """
        return self._rewrap(self.default.encrypt(*keys, **kwargs))

    def decrypt(self, key):
        """
        Decrypt this message with key.

        :param key: The key to decrypt with.
        :type key: pgpy.PGPKey
        :raises: pgpy.errors.PGPError
        :return:
        :rtype: PGPWrapper
        """
        result = None
        if self.mime.is_encrypted():
            result = self.mime.decrypt(key)
        elif self.multisig.is_encrypted():
            result = self.multisig.decrypt(key)
        elif self.inline.is_encrypted():
            result = self.inline.decrypt(key)
        return self._rewrap(result)

    def try_decrypt(self, key):
        """
        Try decrypting the message with given key.

        :param key: The key to decrypt with.
        :type key: pgpy.PGPKey
        :return: The decrypted message, if successfully decrypted,
                 else original message.
        :rtype: PGPWrapper
        """
        try:
            return self._rewrap(self.decrypt(key))
        except PGPError:
            return self

    def sign_encrypt(self, key, *keys, **kwargs):
        """
        Sign and encrypt the message, in one go.

        :param key: The key to sign with.
        :type key: pgpy.PGPKey
        :param keys: The key/s to encrypt with.
        :type keys: pgpy.PGPKey
        :param hash:
        :type hash: pgpy.constants.HashAlgorithm
        :param cipher:
        :type cipher: pgpy.constants.SymmetricKeyAlgorithm
        :return:
        :rtype: PGPWrapper
        """
        return self._rewrap(self.default.sign_encrypt(key, *keys, **kwargs))

    def is_keys(self):
        """
        Whether the message is all keys (all parts).

        :return: If the message is keys.
        :rtype: bool
        """
        return any(wrapper.is_keys() for wrapper in self.wrappers)

    def has_keys(self):
        """
        Whether the message contains public or private keys.

        :return: If the message contains keys.
        :rtype: bool
        """
        return any(wrapper.has_keys() for wrapper in self.wrappers)

    def keys(self):
        """
        Get the collection of keys in this message.

        :return: A collection of keys.
        :rtype: typing.Generator[pgpy.PGPKey]
        """
        if self.mime.has_keys():
            yield from self.mime.keys()
        elif self.multisig.has_keys():
            yield from self.multisig.keys()
        elif self.inline.has_keys():
            yield from self.inline.keys()

    def has_revocs(self):
        """

        :return:
        :rtype: bool
        """
        return any(wrapper.has_revocs() for wrapper in self.wrappers)

    def is_revocs(self):
        """

        :return:
        :rtype: bool
        """
        return any(wrapper.is_revocs() for wrapper in self.wrappers)

    def revocs(self):
        """

        :return:
        :rtype: typing.Generator[pgpy.PGPSignature]
        """
        if self.mime.has_revocs():
            yield from self.mime.revocs()
        elif self.multisig.has_revocs():
            yield from self.multisig.revocs()
        elif self.inline.has_revocs():
            yield from self.inline.revocs()



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

"""Filesystem stored PGP key."""
from os import remove, urandom
from os.path import getmtime, getsize, join
from pathlib import Path

from public import public

from mailman_pgp.utils.file import locked_obj
from mailman_pgp.utils.pgp import key_from_file


@public
class FSKey:
    """Filesystem stored PGP key."""

    def __init__(self, keydir, keyfile, load=False):
        self._key = None
        self._mtime = None
        if isinstance(keydir, Path):
            self.keydir = str(keydir)
        else:
            self.keydir = keydir
        self.keyfile = keyfile
        if load:
            self.load()

    @property
    def key(self):
        """

        :rtype: pgpy.PGPKey
        """
        return self._key

    @key.setter
    def key(self, value):
        """

        :param value:
        :type value: pgpy.PGPKey
        """
        self._key = value

    @property
    def key_path(self):
        return join(self.keydir, self.keyfile)

    @property
    def lock_path(self):
        return self.key_path + '.lock'

    def _load(self):
        try:
            self.key = key_from_file(self.key_path)
            self._mtime = getmtime(self.key_path)
        except FileNotFoundError:
            self.key = None
            self._mtime = None

    @locked_obj('lock_path')
    def load(self):
        self._load()

    @locked_obj('lock_path')
    def reload(self):
        if self.key is None:
            self._load()
        else:
            mtime = getmtime(self.key_path)
            if self._mtime is None or mtime > self._mtime:
                self._load()

    @locked_obj('lock_path')
    def save(self):
        if self.key is None:
            remove(self.key_path)
            self._mtime = None
        else:
            with open(self.key_path, 'w') as key_file:
                key_file.write(str(self.key))
            self._mtime = getmtime(self.key_path)

    @locked_obj('lock_path')
    def delete(self):
        try:
            remove(self.key_path)
        except FileNotFoundError:
            pass

    @locked_obj('lock_path')
    def shred(self):
        try:
            size = getsize(self.key_path)
            for _ in range(50):
                with open(self.key_path, 'wb') as f:
                    data = urandom(size)
                    f.write(data)
            remove(self.key_path)
        except FileNotFoundError:
            pass



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

"""Model for PGP enabled addresses."""

from mailman.database.types import SAUnicode
from mailman.interfaces.usermanager import IUserManager
from public import public
from sqlalchemy import Boolean, Column, Integer, String
from sqlalchemy.orm import reconstructor
from zope.component import getUtility

from mailman_pgp.config import config
from mailman_pgp.model.base import Base
from mailman_pgp.model.fs_key import FSKey


@public
class PGPAddress(Base):
    """A PGP enabled address."""

    __tablename__ = 'pgp_addresses'

    id = Column(Integer, primary_key=True)
    email = Column(SAUnicode, index=True, unique=True)
    key_fingerprint = Column(String(50))
    key_confirmed = Column(Boolean, default=False)

    def __init__(self, address):
        super().__init__(email=address.email)
        self._init()
        self._address = address

    @reconstructor
    def _init(self):
        self._address = None
        self._key = FSKey(config.pgp.keydirs['user_keydir'],
                          self.email + '.asc', True)

    @property
    def key(self):
        """

        :return:
        :rtype: pgpy.PGPKey
        """
        self._key.reload()
        return self._key.key

    @key.setter
    def key(self, value):
        """

        :param value:
        :type value: pgpy.PGPKey
        """
        if value is None:
            self.key_fingerprint = None
        else:
            self.key_fingerprint = value.fingerprint
        self._key.key = value
        self._key.save()

    @property
    def key_path(self):
        """

        :return:
        :rtype: str
        """
        return self._key.key_path

    @property
    def address(self):
        """

        :return:
        :rtype: mailman.model.address.Address
        """
        if self._address is None:
            self._address = getUtility(IUserManager).get_address(self.email)
        return self._address

    @staticmethod
    def for_address(address):
        """

        :param address:
        :return:
        :rtype: PGPAddress|None
        """
        if address is None:
            return None
        return PGPAddress.for_email(address.email)

    @staticmethod
    def for_email(email):
        """

        :param email:
        :return:
        :rtype: PGPAddress|None
        """
        if email is None:
            return None
        return PGPAddress.query().filter_by(email=email).first()



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

""""""
import os
import unittest
from contextlib import ExitStack
from mailbox import Maildir
from tempfile import TemporaryDirectory

from mailman.app.lifecycle import create_list
from mailman.testing.helpers import specialized_message_from_string as mfs

from mailman_pgp.archivers.local_maildir import LocalMaildirArchiver
from mailman_pgp.database import mm_transaction
from mailman_pgp.model.list import PGPMailingList
from mailman_pgp.pgp.mime import MIMEWrapper
from mailman_pgp.testing.config import patch_config
from mailman_pgp.testing.layers import PGPConfigLayer
from mailman_pgp.testing.pgp import load_key


class TestPGPMaildirArchiver(unittest.TestCase):
    layer = PGPConfigLayer

    def setUp(self):
        self.msg = mfs("""\
To: test@example.com
From: anne@example.com
Subject: Testing the test list
Message-ID: <ant>
Message-ID-Hash: MS6QLWERIJLGCRF44J7USBFDELMNT2BW

Tests are better than no tests
but the water deserves to be swum.
""")
        with mm_transaction():
            self.mlist = create_list('test@example.com',
                                     style_name='pgp-default')
        self.pgp_list = PGPMailingList.for_list(self.mlist)
        self.list_key = load_key('ecc_p256.priv.asc')
        self.pgp_list.key = self.list_key

    def test_no_links(self):
        self.assertIsNone(LocalMaildirArchiver.list_url(self.mlist))
        self.assertIsNone(LocalMaildirArchiver.permalink(self.mlist, self.msg))

    def test_no_pgp_list(self):
        with mm_transaction():
            ordinary = create_list('ordinary@example.com')

        LocalMaildirArchiver.archive_message(ordinary, self.msg)

    def test_archives(self):
        with ExitStack() as res:
            maildir_dir = res.enter_context(TemporaryDirectory())
            res.enter_context(patch_config('archiving', 'maildir_dir',
                                           maildir_dir))

            LocalMaildirArchiver.archive_message(self.mlist, self.msg)

            list_dir = os.path.join(maildir_dir, self.mlist.fqdn_listname)

            maildir = Maildir(list_dir)
            messages = maildir.values()
            self.assertEqual(len(messages), 1)

            message = messages[0]
            wrapped = MIMEWrapper(message)
            self.assertTrue(wrapped.is_encrypted())
            decrypted = wrapped.decrypt(self.list_key).msg
            self.assertTrue(self.msg.as_string(), decrypted.as_string())



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

""""""
import unittest

from pkg_resources import resource_filename, resource_string

from mailman_pgp.config import config
from mailman_pgp.config.config import Config
from mailman_pgp.config.validator import ConfigValidator
from mailman_pgp.testing.layers import PGPConfigLayer, PGPLayer


class TestConfig(unittest.TestCase):
    layer = PGPConfigLayer

    def test_name(self):
        self.assertEqual(config.name, 'pgp')


class TestConfigs(unittest.TestCase):
    layer = PGPLayer

    def setUp(self):
        self.validator = ConfigValidator(
                resource_string('mailman_pgp.config',
                                'schema.cfg').decode('utf-8'))

    def test_default_config(self):
        cfg = Config()
        cfg.read(resource_filename('mailman_pgp.config', 'mailman_pgp.cfg'))
        self.validator.validate(cfg)

    def test_testing_config(self):
        cfg = Config()
        cfg.read(resource_filename('mailman_pgp.testing', 'mailman_pgp.cfg'))
        self.validator.validate(cfg)



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

"""Base class for all models."""

from public import public
from sqlalchemy.ext.declarative import as_declarative

from mailman_pgp.config import config


@public
@as_declarative()
class Base:
    """Custom declarative base."""

    @classmethod
    def query(cls):
        """
        A query helper.

        :return: A query on class.
        :rtype: sqlalchemy.orm.query.Query
        """
        return config.db.session.query(cls)



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

"""PGP enabled IndividualDelivery."""
from mailman.mta.base import IndividualDelivery
from mailman.mta.decorating import DecoratingMixin
from mailman.mta.personalized import PersonalizedMixin
from mailman.mta.verp import VERPMixin
from public import public

from mailman_pgp.model.address import PGPAddress
from mailman_pgp.model.list import PGPMailingList
from mailman_pgp.pgp.mime import MIMEWrapper


class PGPIndividualMixin:
    """Individual encryption and signing Delivery mixin."""

    def sign_encrypt(self, mlist, msg, msgdata):
        """
        Sign and encrypt the outgoing message to the recipient.

        :param mlist:
        :type mlist: mailman.model.mailinglist.MailingList
        :param msg:
        :type msg: mailman.email.message.Message
        :param msgdata:
        :type msgdata: dict
        """
        pgp_list = PGPMailingList.for_list(mlist)
        if not pgp_list:
            return
        if not pgp_list.encrypt_outgoing and not pgp_list.sign_outgoing:
            # nothing to do
            return

        recipient = msgdata['recipient']
        pgp_address = PGPAddress.for_email(recipient)
        if pgp_address is None:
            msgdata['no_deliver'] = True
            return
        if pgp_address.key is None or not pgp_address.key_confirmed:
            msgdata['no_deliver'] = True
            return

        key = pgp_address.key
        wrapped = MIMEWrapper(msg)

        if pgp_list.sign_outgoing:
            if pgp_list.encrypt_outgoing:
                wrapped.sign_encrypt(pgp_list.key, key, pgp_list.pubkey)
            else:
                wrapped.sign(pgp_list.key)
        else:
            # Definitely encrypt here, the case where we don't encrypt or sign
            # is handled above at the start of the func.
            wrapped.encrypt(key, pgp_list.pubkey)


@public
class PGPPersonalizedDelivery(IndividualDelivery, VERPMixin, DecoratingMixin,
                              PersonalizedMixin, PGPIndividualMixin):
    """Individual PGP enabled delivery."""

    def __init__(self):
        super().__init__()
        self.callbacks.extend([
            self.avoid_duplicates,
            self.decorate,
            self.personalize_to,
            self.sign_encrypt
        ])

    def _deliver_to_recipients(self, mlist, msg, msgdata, recipients):
        if msgdata.get('no_deliver', False):
            return dict((recipient, (444, BaseException)) for recipient in
                        recipients)
        return super()._deliver_to_recipients(mlist, msg, msgdata, recipients)



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

""""""
import datetime
import time
import unittest

from parameterized import parameterized
from pgpy import PGPKey, PGPUID
from pgpy.constants import (
    CompressionAlgorithm, EllipticCurveOID, HashAlgorithm, KeyFlags,
    PubKeyAlgorithm, SymmetricKeyAlgorithm)

from mailman_pgp.testing.layers import PGPLayer
from mailman_pgp.testing.pgp import load_blob, load_key
from mailman_pgp.utils.pgp import key_usable, revoc_from_blob


class TestPGPUtils(unittest.TestCase):
    layer = PGPLayer

    @parameterized.expand([
        (load_blob('revocs', 'rsa_1024.revoc.asc'),
         load_key('rsa_1024.pub.asc')),
        (load_blob('revocs', 'ecc_secp256k1.revoc.asc'),
         load_key('ecc_secp256k1.pub.asc')),
        (load_blob('revocs', 'ecc_p256.revoc.asc'),
         load_key('ecc_p256.pub.asc'))
    ])
    def test_revoc_from_blob_valid(self, blob, key):
        revoc = revoc_from_blob(blob)
        verifies = key.verify(key, revoc)
        self.assertTrue(bool(verifies))

    @parameterized.expand([
        ('Not an ASCII-Armored blob',),
        (load_blob('keys', 'rsa_1024.pub.asc'),),
    ])
    def test_revoc_from_blob_invalid(self, blob):
        self.assertRaises(ValueError, revoc_from_blob, blob)

    def test_key_usable(self):
        key = load_key('rsa_1024.priv.asc')

        self.assertTrue(key_usable(key.pubkey,
                                   {KeyFlags.Certify, KeyFlags.Sign,
                                    KeyFlags.EncryptCommunications,
                                    KeyFlags.EncryptStorage}))

    def test_key_usable_expired(self):
        key = PGPKey.new(PubKeyAlgorithm.ECDSA, EllipticCurveOID.SECP256K1)
        uid = PGPUID.new('Some Name', email='anne@example.org')
        key.add_uid(uid, key_expiration=datetime.timedelta(seconds=1),
                    usage={KeyFlags.Certify,
                           KeyFlags.Authentication,
                           KeyFlags.Sign},
                    hashes=[HashAlgorithm.SHA256,
                            HashAlgorithm.SHA512],
                    ciphers=[SymmetricKeyAlgorithm.AES256],
                    compression=[CompressionAlgorithm.ZLIB])

        time.sleep(2)

        self.assertFalse(key_usable(key.pubkey, set()))

    def test_key_usable_revoked_uid(self):
        key = load_key('ecc_p256.priv.asc')
        uid = next(iter(key.userids))
        rsig = key.revoke(uid)
        uid |= rsig
        self.assertFalse(key_usable(key.pubkey, {KeyFlags.Sign}))

    def test_key_usable_revoked(self):
        key = load_key('ecc_p256.priv.asc')
        rsig = key.revoke(key)
        key |= rsig

        self.assertFalse(key_usable(key.pubkey, set()))

    def test_key_usable_subkey_revoked(self):
        key = load_key('ecc_p256.priv.asc')
        sub = next(iter(key.subkeys.values()))
        rsig = key.revoke(sub)
        sub |= rsig

        self.assertFalse(
                key_usable(key.pubkey, {KeyFlags.EncryptCommunications}))



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

""""""

import unittest
from contextlib import suppress
from unittest.mock import patch

from mailman.app.lifecycle import create_list
from mailman.interfaces.pending import IPendings
from mailman.interfaces.usermanager import IUserManager
from mailman.interfaces.workflows import IWorkflow
from mailman.testing.helpers import get_queue_messages
from mailman.workflows.common import SubscriptionBase
from zope.component import getUtility
from zope.interface import implementer

from mailman_pgp.database import mm_transaction, transaction
from mailman_pgp.model.address import PGPAddress
from mailman_pgp.model.list import PGPMailingList
from mailman_pgp.pgp.wrapper import PGPWrapper
from mailman_pgp.testing.layers import PGPConfigLayer
from mailman_pgp.testing.pgp import load_key
from mailman_pgp.workflows.base import (PGPMixin)
from mailman_pgp.workflows.key_confirm import ConfirmPubkeyMixin
from mailman_pgp.workflows.key_set import KEY_REQUEST, SetPubkeyMixin


class PubkeyMixinTestSetup():
    def setUp(self):
        with mm_transaction():
            self.mlist = create_list('test@example.com',
                                     style_name='pgp-default')

        self.list_key = load_key('ecc_p256.priv.asc')

        with transaction():
            self.pgp_list = PGPMailingList.for_list(self.mlist)
            self.pgp_list.key = self.list_key

        self.um = getUtility(IUserManager)

        self.sender_key = load_key('rsa_1024.priv.asc')
        self.sender = self.um.create_address('anne@example.org')


@implementer(IWorkflow)
class PGPTestWorkflow(SubscriptionBase, PGPMixin, SetPubkeyMixin,
                      ConfirmPubkeyMixin):
    name = 'test-workflow'
    description = ''
    initial_state = 'prepare'
    save_attributes = (
        'pubkey_key',
        'pubkey_confirmed',
        'address_key',
        'subscriber_key',
        'user_key',
        'token_owner_key'
    )

    def __init__(self, mlist, subscriber=None, *, pubkey=None,
                 pubkey_pre_confirmed=False):
        SubscriptionBase.__init__(self, mlist, subscriber)
        SetPubkeyMixin.__init__(self, pubkey=pubkey)
        ConfirmPubkeyMixin.__init__(self, pre_confirmed=pubkey_pre_confirmed)
        PGPMixin.__init__(self, mlist)

    def _step_prepare(self):
        self.push('do_subscription')
        self.push('pubkey_confirmation')
        self.push('pubkey_checks')
        self.push('create_address')
        self.push('sanity_checks')


class TestPGPMixin(PubkeyMixinTestSetup, unittest.TestCase):
    layer = PGPConfigLayer

    def test_create_address(self):
        workflow = PGPTestWorkflow(self.mlist, self.sender)
        workflow.run_thru('create_address')
        pgp_address = PGPAddress.for_address(self.sender)
        self.assertIsNotNone(pgp_address)

    def test_address_existing(self):
        workflow = PGPTestWorkflow(self.mlist, self.sender)
        with transaction() as t:
            pgp_address = PGPAddress(self.sender)
            t.add(pgp_address)
        workflow.run_thru('create_address')
        still = PGPAddress.for_address(self.sender)
        self.assertIsNotNone(still)


class TestSetPubkeyMixin(PubkeyMixinTestSetup, unittest.TestCase):
    layer = PGPConfigLayer

    def test_key_request_sent(self):
        workflow = PGPTestWorkflow(self.mlist, self.sender)
        list(workflow)
        items = get_queue_messages('virgin', expected_count=1)
        message = items[0].msg
        token = workflow.token

        self.assertEqual(message['Subject'], 'key set {}'.format(token))
        self.assertEqual(message.get_payload(), KEY_REQUEST)

    def test_receive_key(self):
        workflow = PGPTestWorkflow(self.mlist, self.sender)
        list(workflow)
        with transaction():
            pgp_address = PGPAddress.for_address(self.sender)
            pgp_address.key = self.sender_key.pubkey

        receive_workflow = PGPTestWorkflow(self.mlist)
        receive_workflow.token = workflow.token
        receive_workflow.restore()
        receive_workflow.run_thru('receive_key')

    def test_set_pubkey(self):
        workflow = PGPTestWorkflow(self.mlist, self.sender,
                                   pubkey=self.sender_key.pubkey)
        workflow.run_thru('pubkey_checks')
        pgp_address = PGPAddress.for_address(self.sender)
        self.assertIsNotNone(pgp_address)
        self.assertIsNotNone(pgp_address.key)
        self.assertEqual(pgp_address.key_fingerprint,
                         self.sender_key.fingerprint)

    def test_pubkey_set(self):
        workflow = PGPTestWorkflow(self.mlist, self.sender)
        with transaction() as t:
            pgp_address = PGPAddress(self.sender)
            pgp_address.key = self.sender_key.pubkey
            t.add(pgp_address)
        workflow.run_thru('pubkey_checks')
        self.assertEqual(pgp_address.key_fingerprint,
                         self.sender_key.fingerprint)


class TestConfirmPubkeyMixin(PubkeyMixinTestSetup, unittest.TestCase):
    layer = PGPConfigLayer

    def test_key_request_pubkey_set(self):
        workflow = PGPTestWorkflow(self.mlist, self.sender,
                                   pubkey=self.sender_key.pubkey,
                                   pubkey_pre_confirmed=True)
        workflow.run_thru('pubkey_confirmation')
        with patch.object(workflow, '_step_do_subscription') as step:
            next(workflow)
        step.assert_called_once_with()

    def test_send_key_confirm_request(self):
        workflow = PGPTestWorkflow(self.mlist, self.sender,
                                   pubkey=self.sender_key.pubkey,
                                   pubkey_pre_confirmed=False)
        list(workflow)
        items = get_queue_messages('virgin', expected_count=1)
        message = items[0].msg
        token = workflow.token

        self.assertEqual(message['Subject'], 'key confirm {}'.format(token))
        wrapped = PGPWrapper(message)
        self.assertTrue(wrapped.is_encrypted())

    def test_receive_confirmation(self):
        workflow = PGPTestWorkflow(self.mlist, self.sender,
                                   pubkey=self.sender_key.pubkey,
                                   pubkey_pre_confirmed=False)
        list(workflow)

        receive_workflow = PGPTestWorkflow(self.mlist)
        receive_workflow.token = workflow.token
        receive_workflow.restore()
        receive_workflow.run_thru('receive_key_confirmation')
        with patch.object(receive_workflow, '_step_do_subscription') as step:
            next(receive_workflow)
        step.assert_called_once_with()


class TestBothPubkeyMixins(PubkeyMixinTestSetup, unittest.TestCase):
    layer = PGPConfigLayer

    def test_pended_data_key_request(self):
        workflow = PGPTestWorkflow(self.mlist, self.sender)
        with suppress(StopIteration):
            workflow.run_thru('send_key_request')
        self.assertIsNotNone(workflow.token)
        pendable = getUtility(IPendings).confirm(workflow.token, expunge=False)
        self.assertEqual(pendable['list_id'], 'test.example.com')
        self.assertEqual(pendable['email'], 'anne@example.org')
        self.assertEqual(pendable['display_name'], '')
        self.assertEqual(pendable['when'], '2005-08-01T07:49:23')
        self.assertEqual(pendable['token_owner'], 'subscriber')

    def test_pended_data_key_confirmation(self):
        workflow = PGPTestWorkflow(self.mlist, self.sender,
                                   pubkey=self.sender_key.pubkey)
        with suppress(StopIteration):
            workflow.run_thru('send_key_confirm_request')
        self.assertIsNotNone(workflow.token)
        pendable = getUtility(IPendings).confirm(workflow.token, expunge=False)
        self.assertEqual(pendable['list_id'], 'test.example.com')
        self.assertEqual(pendable['email'], 'anne@example.org')
        self.assertEqual(pendable['display_name'], '')
        self.assertEqual(pendable['when'], '2005-08-01T07:49:23')
        self.assertEqual(pendable['token_owner'], 'subscriber')

    def test_exisitng_pgp_address(self):
        workflow = PGPTestWorkflow(self.mlist, self.sender)

        with transaction() as t:
            pgp_address = PGPAddress(self.sender)
            pgp_address.key = self.sender_key.pubkey
            pgp_address.key_confirmed = True
            t.add(pgp_address)

        workflow.run_thru('pubkey_confirmation')
        with patch.object(workflow, '_step_do_subscription') as step:
            next(workflow)
        step.assert_called_once_with()

    def test_exisitng_pgp_address_not_confirmed(self):
        workflow = PGPTestWorkflow(self.mlist, self.sender)

        with transaction() as t:
            pgp_address = PGPAddress(self.sender)
            pgp_address.key = self.sender_key.pubkey
            t.add(pgp_address)

        workflow.run_thru('pubkey_confirmation')
        with patch.object(workflow, '_step_send_key_confirm_request') as step:
            next(workflow)
        step.assert_called_once_with()

    def test_exisitng_pgp_address_no_key(self):
        workflow = PGPTestWorkflow(self.mlist, self.sender)

        with transaction() as t:
            pgp_address = PGPAddress(self.sender)
            t.add(pgp_address)

        workflow.run_thru('pubkey_checks')
        with patch.object(workflow, '_step_send_key_request') as step:
            next(workflow)
        step.assert_called_once_with()



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

""""""
from public import public
from sqlalchemy import Column, LargeBinary, String

from mailman_pgp.model.base import Base


@public
class PGPSigHash(Base):
    """"""

    __tablename__ = 'pgp_sighashes'

    hash = Column(LargeBinary, primary_key=True)
    fingerprint = Column(String(50), index=True)

    @staticmethod
    def hashes(hashes):
        if hashes is None or len(hashes) == 0:
            return None
        return PGPSigHash.query().filter(PGPSigHash.hash.in_(hashes)).all()



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

"""REST root."""

from mailman.rest.helpers import child, NotFound
from public import public

from mailman_pgp.rest.addresses import AllAddresses, AnAddress
from mailman_pgp.rest.config import AConfig
from mailman_pgp.rest.lists import AllPGPLists, APGPList


@public
class RESTRoot:
    @child()
    def lists(self, context, segments):
        if len(segments) == 0:
            return AllPGPLists(), []
        else:
            list_identifier = segments.pop(0)
            return APGPList(list_identifier), segments

    @child()
    def addresses(self, context, segments):
        if len(segments) == 0:
            return AllAddresses(), []
        else:
            email = segments.pop(0)
            return AnAddress(email), segments

    @child()
    def config(self, context, segments):
        if len(segments) == 0:
            return AConfig(), []
        elif len(segments) == 1:
            section = segments.pop(0)
            return AConfig(section), []
        elif len(segments) == 2:
            section = segments.pop(0)
            attribute = segments.pop(0)
            return AConfig(section, attribute), []
        else:
            return NotFound(), []



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

""""""
import logging
import time

from mailman.config import config
from mailman.interfaces.mailinglist import Personalization
from mailman.interfaces.mta import SomeRecipientsFailed
from mailman.mta.bulk import BulkDelivery
from mailman.mta.deliver import Deliver
from mailman.utilities.string import expand
from public import public

from mailman_pgp.model.list import PGPMailingList
from mailman_pgp.mta.bulk import PGPBulkDelivery
from mailman_pgp.mta.personalized import PGPPersonalizedDelivery

COMMA = ','
log = logging.getLogger('mailman.smtp')


@public
def deliver(mlist, msg, msgdata):
    """Deliver a message to the outgoing mail server."""
    # If there are no recipients, there's nothing to do.
    recipients = msgdata.get('recipients')
    if not recipients:
        # Could be None, could be an empty sequence.
        return
    # Which delivery agent should we use?  Several situations can cause us to
    # use individual delivery.  If not specified, use bulk delivery.  See the
    # to-outgoing handler for when the 'verp' key is set in the metadata.
    personalized_agent = Deliver
    bulk_agent = BulkDelivery

    pgp_list = PGPMailingList.for_list(mlist)
    if pgp_list and msgdata.get('pgp_is_posting', False):
        personalized_agent = PGPPersonalizedDelivery
        bulk_agent = PGPBulkDelivery

    if msgdata.get('verp', False):
        agent = personalized_agent()
    elif mlist.personalize != Personalization.none:
        agent = personalized_agent()
    else:
        agent = bulk_agent(int(config.mta.max_recipients))
    log.debug('Using agent: %s', agent)
    # Keep track of the original recipients and the original sender for
    # logging purposes.
    original_recipients = msgdata['recipients']
    original_sender = msgdata.get('original-sender', msg.sender)
    # Let the agent attempt to deliver to the recipients.  Record all failures
    # for re-delivery later.
    t0 = time.time()
    refused = agent.deliver(mlist, msg, msgdata)
    t1 = time.time()
    # Log this posting.
    size = getattr(msg, 'original_size', msgdata.get('original_size'))
    if size is None:
        size = len(msg.as_string())
    substitutions = dict(
            msgid=msg.get('message-id', 'n/a'),  # noqa: E221, E251
            listname=mlist.fqdn_listname,  # noqa: E221, E251
            sender=original_sender,  # noqa: E221, E251
            recip=len(original_recipients),  # noqa: E221, E251
            size=size,  # noqa: E221, E251
            time=t1 - t0,  # noqa: E221, E251
            refused=len(refused),  # noqa: E221, E251
            smtpcode='n/a',  # noqa: E221, E251
            smtpmsg='n/a',  # noqa: E221, E251
    )
    template = config.logging.smtp.every
    if template.lower() != 'no':
        log.info('%s', expand(template, mlist, substitutions))
    if refused:
        template = config.logging.smtp.refused
        if template.lower() != 'no':
            log.info('%s', expand(template, mlist, substitutions))
    else:
        # Log the successful post, but if it was not destined to the mailing
        # list (e.g. to the owner or admin), print the actual recipients
        # instead of just the number.
        if not msgdata.get('tolist', False):
            recips = msg.get_all('to', [])
            recips.extend(msg.get_all('cc', []))
            substitutions['recips'] = COMMA.join(recips)
        template = config.logging.smtp.success
        if template.lower() != 'no':
            log.info('%s', expand(template, mlist, substitutions))
    # Process any failed deliveries.
    temporary_failures = []
    permanent_failures = []
    for recipient, (code, smtp_message) in refused.items():
        # RFC 5321, $4.5.3.1.10 says:
        #
        #   RFC 821 [1] incorrectly listed the error where an SMTP server
        #   exhausts its implementation limit on the number of RCPT commands
        #   ("too many recipients") as having reply code 552.  The correct
        #   reply code for this condition is 452.  Clients SHOULD treat a 552
        #   code in this case as a temporary, rather than permanent, failure
        #   so the logic below works.
        #
        if code >= 500 and code != 552:
            # A permanent failure
            permanent_failures.append(recipient)
        else:
            # Deal with persistent transient failures by queuing them up for
            # future delivery.  TBD: this could generate lots of log entries!
            temporary_failures.append(recipient)
        template = config.logging.smtp.failure
        if template.lower() != 'no':
            substitutions.update(
                    recip=recipient,  # noqa: E221, E251
                    smtpcode=code,  # noqa: E221, E251
                    smtpmsg=smtp_message  # noqa: E221, E251
            )
            log.info('%s', expand(template, mlist, substitutions))
    # Return the results
    if temporary_failures or permanent_failures:
        raise SomeRecipientsFailed(temporary_failures, permanent_failures)



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

""""""

import unittest

from mailman.app.lifecycle import create_list
from mailman.interfaces.subscriptions import ISubscriptionManager, TokenOwner
from mailman.interfaces.usermanager import IUserManager
from mailman.testing.helpers import get_queue_messages
from zope.component import getUtility

from mailman_pgp.config import mm_config
from mailman_pgp.database import mm_transaction, transaction
from mailman_pgp.model.address import PGPAddress
from mailman_pgp.model.list import PGPMailingList
from mailman_pgp.pgp.wrapper import PGPWrapper
from mailman_pgp.testing.layers import PGPConfigLayer
from mailman_pgp.testing.pgp import load_key
from mailman_pgp.workflows.key_change import (KeyChangeModWorkflow,
                                              KeyChangeWorkflow)


class TestKeyChangeWorkflow(unittest.TestCase):
    layer = PGPConfigLayer

    def setUp(self):
        with mm_transaction():
            self.mlist = create_list('test@example.com',
                                     style_name='pgp-default')
        with transaction():
            self.pgp_list = PGPMailingList.for_list(self.mlist)
            self.pgp_list.key = load_key('ecc_p256.priv.asc')

        self.sender_key = load_key('rsa_1024.priv.asc')
        self.sender_new_key = load_key('ecc_p256.priv.asc')
        self.sender = getUtility(IUserManager).create_address(
                'anne@example.org')

    def test_has_workflows(self):
        self.assertTrue(KeyChangeWorkflow.name, mm_config.workflows)
        self.assertTrue(KeyChangeModWorkflow.name, mm_config.workflows)

    def test_pgp_address_none(self):
        workflow = KeyChangeWorkflow(self.mlist)
        with self.assertRaises(ValueError):
            list(workflow)

    def test_pubkey_none(self):
        with transaction() as t:
            pgp_address = PGPAddress(self.sender)
            t.add(pgp_address)

        workflow = KeyChangeWorkflow(self.mlist, pgp_address)
        with self.assertRaises(ValueError):
            list(workflow)

    def test_send_key_confirm_request(self):
        with transaction() as t:
            pgp_address = PGPAddress(self.sender)
            pgp_address.key = self.sender_key.pubkey
            pgp_address.key_confirmed = True
            t.add(pgp_address)

        workflow = KeyChangeWorkflow(self.mlist, pgp_address,
                                     self.sender_new_key.pubkey)
        list(workflow)
        items = get_queue_messages('virgin', expected_count=1)
        message = items[0].msg
        token = workflow.token

        self.assertEqual(message['Subject'], 'key confirm {}'.format(token))
        wrapped = PGPWrapper(message)
        self.assertTrue(wrapped.is_encrypted())

    def test_confirm(self):
        with transaction() as t:
            pgp_address = PGPAddress(self.sender)
            pgp_address.key = self.sender_key.pubkey
            pgp_address.key_confirmed = True
            t.add(pgp_address)

        workflow = KeyChangeWorkflow(self.mlist, pgp_address,
                                     self.sender_new_key.pubkey)
        list(workflow)

        token, token_owner, member = ISubscriptionManager(self.mlist).confirm(
                workflow.token)
        self.assertIsNone(token)
        self.assertEqual(token_owner, TokenOwner.no_one)

        pgp_address = PGPAddress.for_address(self.sender)
        self.assertEqual(pgp_address.key_fingerprint,
                         self.sender_new_key.fingerprint)
        self.assertTrue(pgp_address.key_confirmed)

    def test_confirm_mod(self):
        with transaction() as t:
            pgp_address = PGPAddress(self.sender)
            pgp_address.key = self.sender_key.pubkey
            pgp_address.key_confirmed = True
            t.add(pgp_address)

        workflow = KeyChangeModWorkflow(self.mlist, pgp_address,
                                        self.sender_new_key.pubkey)
        list(workflow)

        token, token_owner, member = ISubscriptionManager(self.mlist).confirm(
                workflow.token)
        self.assertIsNotNone(token)
        self.assertEqual(token_owner, TokenOwner.moderator)



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
import ast


class enumflag_validator:
    def __init__(self, enum):
        self.enum = enum

    def __call__(self, value):
        if isinstance(value, (tuple, list)):
            result = set()
            for val in value:
                result.add(self.enum[val])
            return result
        if value in self.enum:
            return {self.enum[value]}
        else:
            l = ast.literal_eval(value)
            if isinstance(l, (list, tuple)):
                result = set()
                for val in l:
                    result.add(self.enum[val])
                return result
            raise ValueError


class workflow_validator:
    def __init__(self, *classes):
        self.classes = classes

    def __call__(self, value):
        if value in (workflow.name for workflow in self.classes):
            return value
        raise ValueError
